# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Dataclasses models to generate table rows."""

from datetime import datetime
from uuid import UUID

from sqlalchemy.orm import Mapped, mapped_column

from ...load.postgresql.models import Model


class PersistentIdentifier(Model):
    """Persistent identifier dataclass model."""

    __tablename__ = "pidstore_pid"

    id: Mapped[int] = mapped_column(primary_key=True)
    pid_type: Mapped[str]
    pid_value: Mapped[str]
    status: Mapped[str]
    object_type: Mapped[str] = mapped_column(nullable=True)
    object_uuid: Mapped[UUID] = mapped_column(nullable=True)
    created: Mapped[datetime]
    updated: Mapped[datetime]
    # default since we do not take it into account in many places
    # ant it's not in the data
    pid_provider: Mapped[str] = mapped_column(nullable=True, default=None)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Communities table generator tests."""


from invenio_rdm_migrator.streams.communities.load import CommunityTableGenerator
from invenio_rdm_migrator.streams.models.communities import (
    Community,
    CommunityFile,
    CommunityMember,
)
from invenio_rdm_migrator.streams.models.files import FilesBucket, FilesObjectVersion
from invenio_rdm_migrator.streams.models.oai import OAISet


def test_generate_rows(state, transformed_community_entry_pks):
    """Test the row generation of the request table generator."""
    tg = CommunityTableGenerator()
    rows = list(tg._generate_rows(transformed_community_entry_pks))
    expected_rows = [
        FilesBucket(
            id="12345678-abcd-1a2b-3c4d-123abc456def",
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            default_location=None,
            default_storage_class="L",
            size=0,
            quota_size=0,
            max_file_size=0,
            locked=False,
            deleted=False,
        ),
        Community(
            id="7357c033-abcd-1a2b-3c4d-123abc456def",
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            json={
                "title": "Migrator community",
                "description": "Migrator testing community",
                "page": "",
                "curation_policy": "",
            },
            version_id=1,
            slug="migrator",
            bucket_id="12345678-abcd-1a2b-3c4d-123abc456def",
            deletion_status="P",
        ),
        OAISet(
            id=1,
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            spec="community-migrator",
            name="Migrator community",
            description="Migrator testing community",
            search_pattern="parent.communities.ids:7357c033-abcd-1a2b-3c4d-123abc456def",
            system_created=True,
        ),
        CommunityMember(
            id="12345678-abcd-1a2b-3c4d-123abc456def",
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            json={},
            version_id=1,
            role="owner",
            visible=True,
            active=True,
            community_id="7357c033-abcd-1a2b-3c4d-123abc456def",
            user_id=1,
            group_id=None,
            request_id=None,
        ),
        FilesObjectVersion(
            version_id="12345678-abcd-1a2b-3c4d-123abc456def",
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            key="logo",
            bucket_id="12345678-abcd-1a2b-3c4d-123abc456def",
            file_id="12345678-abcd-1a2b-3c4d-123abc456def",
            _mimetype=None,
            is_head=True,
        ),
        CommunityFile(
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            id="12345678-abcd-1a2b-3c4d-123abc456def",
            json={},
            version_id=1,
            key="logo",
            record_id="7357c033-abcd-1a2b-3c4d-123abc456def",
            object_version_id="12345678-abcd-1a2b-3c4d-123abc456def",
        ),
    ]

    assert rows == expected_rows



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Draft actions tests."""

import pytest

from invenio_rdm_migrator.load.postgresql.transactions.operations import OperationType
from invenio_rdm_migrator.streams.actions.load import (
    DraftCreateAction,
    DraftEditAction,
    DraftPublishNewAction,
)
from invenio_rdm_migrator.streams.models.files import FilesBucket
from invenio_rdm_migrator.streams.models.pids import PersistentIdentifier
from invenio_rdm_migrator.streams.models.records import (
    RDMDraftFile,
    RDMDraftMetadata,
    RDMParentMetadata,
    RDMRecordFile,
    RDMRecordMetadata,
    RDMVersionState,
)


@pytest.mark.skip()
def test_create_draft_new(
    session,
    database,
    draft_data,
    parent_data,
    bucket_data,
    draft_pid_data,
    parent_pid_data,
):
    data = dict(
        draft_pid=draft_pid_data,
        draft_bucket=bucket_data,
        draft=draft_data,
        parent_pid=parent_pid_data,
        parent=parent_data,
    )
    action = DraftCreateAction(data)
    rows = list(action.prepare(session))
    assert len(rows) == 7
    assert rows[0].type == OperationType.INSERT
    assert rows[0].model == PersistentIdentifier
    assert rows[1].type == OperationType.INSERT
    assert rows[1].model == FilesBucket
    assert rows[2].type == OperationType.INSERT
    assert rows[2].model == PersistentIdentifier
    assert rows[3].type == OperationType.INSERT
    assert rows[3].model == RDMParentMetadata
    assert rows[4].type == OperationType.UPDATE
    assert rows[4].model == PersistentIdentifier
    assert rows[5].type == OperationType.INSERT
    assert rows[5].model == RDMDraftMetadata
    assert rows[6].type == OperationType.INSERT
    assert rows[6].model == RDMVersionState


@pytest.mark.skip()
def test_create_draft_new_version(
    session,
    state,
    draft_data,
    parent_data,
    bucket_data,
    draft_pid_data,
    parent_pid_data,
):
    # set existing parent so the action goes on the new version path
    state.PARENTS.add(
        parent_data["json"]["id"],  # recid
        {
            "id": parent_data["id"],
            "latest_id": "93c09d1d-47d2-48e2-9867-ca597b4ebb41",
            "latest_index": 1,
        },
    )
    data = dict(
        draft_pid=draft_pid_data,
        draft_bucket=bucket_data,
        draft=draft_data,
        parent_pid=parent_pid_data,
        parent=parent_data,
    )
    action = DraftCreateAction(data)

    rows = list(action.prepare(session))
    assert len(rows) == 5
    assert rows[0].type == OperationType.INSERT
    assert rows[0].model == PersistentIdentifier
    assert rows[1].type == OperationType.INSERT
    assert rows[1].model == FilesBucket
    assert rows[2].type == OperationType.UPDATE
    assert rows[2].model == PersistentIdentifier
    assert rows[3].type == OperationType.INSERT
    assert rows[3].model == RDMDraftMetadata
    assert rows[4].type == OperationType.INSERT
    assert rows[4].model == RDMVersionState


@pytest.mark.skip()
def test_create_draft_published_of_record(
    session,
    state,
    draft_data,
    parent_data,
    bucket_data,
    draft_pid_data,
    parent_pid_data,
):
    # set existing parent so the action goes on the new version path
    state.PARENTS.add(
        parent_data["json"]["id"],  # recid
        {
            "id": parent_data["id"],
            "latest_id": "93c09d1d-47d2-48e2-9867-ca597b4ebb41",
            "latest_index": 1,
        },
    )

    state.RECORDS.add(
        draft_data["json"]["id"],  # recid
        {
            "index": 1,
            "id": draft_data["id"],  # uuid
            "parent_id": parent_data["id"],  # parent uuid
            "fork_version_id": 1,
            "pids": {},
        },
    )
    data = dict(
        draft_pid=draft_pid_data,
        draft_bucket=bucket_data,
        draft=draft_data,
        parent_pid=parent_pid_data,
        parent=parent_data,
    )
    action = DraftCreateAction(data)
    rows = list(action.prepare(session))
    assert len(rows) == 4
    assert rows[0].type == OperationType.INSERT
    assert rows[0].model == PersistentIdentifier
    assert rows[1].type == OperationType.INSERT
    assert rows[1].model == FilesBucket
    assert rows[2].type == OperationType.INSERT
    assert rows[2].model == RDMDraftMetadata
    assert rows[3].type == OperationType.UPDATE
    assert rows[3].model == RDMVersionState


@pytest.mark.skip()
def test_edit_draft(session, state, draft_data, parent_data):
    # draft id must be on the cache, set there by the draft create action
    parent_state = {
        "id": parent_data["id"],
        "next_draft_id": draft_data["id"],
    }
    state.PARENTS.add(parent_data["json"]["id"], parent_state)
    # the transform will bring in and empty id for both parent and draft
    del draft_data["id"]
    del parent_data["id"]

    data = dict(
        draft=draft_data,
        parent=parent_data,
    )
    action = DraftEditAction(data)
    rows = list(action.prepare(session))
    assert len(rows) == 2
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == RDMParentMetadata
    assert rows[0].data == {"id": parent_state["id"], **parent_data}
    assert rows[1].type == OperationType.UPDATE
    assert rows[1].model == RDMDraftMetadata
    assert rows[1].data == {"id": parent_state["next_draft_id"], **draft_data}


@pytest.mark.skip()
def test_partial_edit_draft(session, state, draft_data, parent_data):
    # draft id must be on the cache, set there by the draft create action
    parent_state = {
        "id": parent_data["id"],
        "next_draft_id": draft_data["id"],
    }
    state.PARENTS.add(parent_data["json"]["id"], parent_state)
    # the transform will bring in and empty id for both parent and draft
    del draft_data["id"]
    del parent_data["id"]

    # remove some content to simulate partial updates
    for key in ["bucket_id", "parent_id", "expires_at", "created", "fork_version_id"]:
        del draft_data[key]

    for key in ["created", "version_id"]:
        del parent_data[key]

    data = dict(
        draft=draft_data,
        parent=parent_data,
    )
    action = DraftEditAction(data)
    rows = list(action.prepare(session))
    assert len(rows) == 2
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == RDMParentMetadata
    assert set(rows[0].data.keys()) == {"id", "json", "updated"}
    assert rows[1].type == OperationType.UPDATE
    assert rows[1].model == RDMDraftMetadata
    assert set(rows[1].data.keys()) == {"id", "json", "version_id", "index", "updated"}


@pytest.mark.skip()
def test_publish_new_draft_with_all_pids(
    session,
    state,
    record_oai_data,
    draft_pid_data,
    draft_doi_data,
    draft_data,
    parent_pid_data,
    parent_doi_data,
    parent_data,
    bucket_data,
    ov_data,  # needed to calculate file records
):
    # pid should be in the state with status reserved so it can be updated (registered)
    state.PIDS.add(
        draft_pid_data["pid_value"],
        {
            "id": draft_pid_data["id"],
            "pid_type": draft_pid_data["pid_type"],
            "status": draft_pid_data["status"],
            "created": draft_pid_data["created"],
            "obj_type": draft_pid_data["object_type"],
        },
    )
    # bucket should be linked to the draft so it can be deleted
    state.BUCKETS.add(bucket_data["id"], {"draft_id": draft_data["id"]})
    # file records should be in the state (e.g. 2 of them)
    state.FILE_RECORDS.add(
        "123b243e-9c0c-44df-bd1f-6decc374cf78",
        {
            "json": {},
            "created": ov_data["created"],
            "updated": ov_data["updated"],
            "version_id": 1,
            "key": ov_data["key"] + "-1",
            "record_id": draft_data["id"],
            "object_version_id": ov_data["version_id"],
        },
    )

    state.FILE_RECORDS.add(
        "124b243e-9c0c-44df-bd1f-6decc374cf78",
        {
            "json": {},
            "created": ov_data["created"],
            "updated": ov_data["updated"],
            "version_id": 1,
            "key": ov_data["key"] + "-2",
            "record_id": draft_data["id"],
            "object_version_id": ov_data["version_id"],
        },
    )

    # parent should be in the state so it can be updated
    state.PARENTS.add(
        parent_pid_data["pid_value"],
        {"id": parent_data["id"], "next_draft_id": draft_data["id"]},
    )

    # update with the data that should be processed (K was for the state)
    draft_pid_data["status"] = parent_pid_data["status"] = "R"
    bucket_data["locked"] = True
    # TEST!
    data = dict(
        bucket=bucket_data,
        parent=parent_data,
        draft=draft_data,
        parent_pid=parent_pid_data,
        parent_doi=parent_doi_data,
        draft_pid=draft_pid_data,
        draft_oai=record_oai_data,
        draft_doi=draft_doi_data,
    )
    action = DraftPublishNewAction(data)
    rows = list(action.prepare(session))
    assert (
        len(rows) == 5 + 1 + 4 + 4
    )  # pids + bucket + file records + draft, record, parent, versioning

    # pids
    assert rows[0].type == OperationType.UPDATE  # recid
    assert rows[0].model == PersistentIdentifier
    assert rows[1].type == OperationType.UPDATE  # parent recid
    assert rows[1].model == PersistentIdentifier
    assert rows[2].type == OperationType.INSERT  # doi
    assert rows[2].model == PersistentIdentifier
    assert rows[3].type == OperationType.INSERT  # parent doi
    assert rows[3].model == PersistentIdentifier
    assert rows[4].type == OperationType.INSERT  # oai
    assert rows[4].model == PersistentIdentifier
    # bucket
    assert rows[5].type == OperationType.UPDATE
    assert rows[5].model == FilesBucket
    # file records
    assert rows[6].type == OperationType.DELETE
    assert rows[6].model == RDMDraftFile
    assert rows[7].type == OperationType.INSERT
    assert rows[7].model == RDMRecordFile
    assert rows[8].type == OperationType.DELETE
    assert rows[8].model == RDMDraftFile
    assert rows[9].type == OperationType.INSERT
    assert rows[9].model == RDMRecordFile
    # draft, record, parent, versioning
    assert rows[10].type == OperationType.DELETE
    assert rows[10].model == RDMDraftMetadata
    assert rows[11].type == OperationType.INSERT
    assert rows[11].model == RDMRecordMetadata
    assert rows[12].type == OperationType.UPDATE
    assert rows[12].model == RDMParentMetadata
    assert rows[13].type == OperationType.UPDATE
    assert rows[13].model == RDMVersionState



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration record transform interfaces."""

from ...transform import IdentityTransform


class OAuthServerScopesMapMixin:
    """OAuth server token data transformation."""

    SCOPES_MAPPING = {
        "deposit:actions": None,
        "deposit:extra_formats": None,
        "deposit:write": None,
        "generate": "tokens:generate",
        "user:email": "user:email",
        "webhooks:event": "webhooks:event",
    }
    """Keys are Invenio v3 scopes and values are new RDM scopes."""

    def map_scopes(self, scopes):
        """Token scopes."""
        scopes = scopes or ""

        if scopes:  # account for cases where scopes = "", otherwise scopes = [""] fails
            scopes = scopes.split(" ")
        else:
            return ""

        new_scopes = []
        for scope in scopes:
            new_scope = self.SCOPES_MAPPING[scope]
            if new_scope:
                new_scopes.append(new_scope)

        return " ".join(new_scopes)


class OAuthServerTokenTransform(IdentityTransform, OAuthServerScopesMapMixin):
    """Transform OAuth server token."""

    def _transform(self, entry):
        """Transform a single entry."""
        data = super()._transform(entry)
        scopes = data.get("_scopes")
        if scopes:
            data["_scopes"] = self.map_scopes(scopes)
        return data


class OAuthServerClientTransform(IdentityTransform, OAuthServerScopesMapMixin):
    """Transform OAuth server client."""

    def _transform(self, entry):
        """Transform a single entry."""
        data = super()._transform(entry)
        scopes = data.get("_default_scopes")
        if scopes:
            data["_default_scopes"] = self.map_scopes(scopes)
        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration record load module."""

from ...load.postgresql.bulk import PostgreSQLCopyLoad
from .table_generators import (
    RDMDeletedRecordTableGenerator,
    RDMDraftTableGenerator,
    RDMRecordTableGenerator,
    RDMVersionStateTableGenerator,
)


class RDMRecordCopyLoad(PostgreSQLCopyLoad):
    """RDM record copy load."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(table_generators=[RDMRecordTableGenerator()], **kwargs)


class RDMDraftCopyLoad(PostgreSQLCopyLoad):
    """RDM draft copy load."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(table_generators=[RDMDraftTableGenerator()], **kwargs)


class RDMVersionStateCopyLoad(PostgreSQLCopyLoad):
    """RDM Vesion state copy load."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(table_generators=[RDMVersionStateTableGenerator()], **kwargs)


class RDMDeletedRecordCopyLoad(PostgreSQLCopyLoad):
    """RDM deleted record copy load."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(table_generators=[RDMDeletedRecordTableGenerator()], **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Single table generator module."""

from .table import TableGenerator


class SingleTableGenerator(TableGenerator):
    """Contains only one table to yield."""

    def __init__(self, table, pks=None, post_load_hooks=None):
        """Constructor."""
        assert not isinstance(table, (list, dict))

        super().__init__(tables=[table], pks=pks, post_load_hooks=post_load_hooks)

    def _generate_rows(self, data, **kwargs):
        """Yield generated rows."""
        table = self.tables[0]
        yield table(**data)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""PostgreSQL Transaction load operations."""

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Type

from ..models import Model


class OperationType(str, Enum):
    """SQL operation type enumeration."""

    INSERT = "C"
    UPDATE = "U"
    DELETE = "D"

    def __eq__(self, other):
        """Equality test."""
        return self.value == other.upper()

    def __str__(self):
        """Return its value."""
        return self.value

    def __repr__(self):
        """Eval-friendly enum representation."""
        return f"{self.__class__.__name__}.{self.name}"


def to_db_type(col, val):
    """Convert value to the appropriate DB column type."""
    if issubclass(col.type.python_type, (datetime,)) and isinstance(val, int):
        return datetime.utcfromtimestamp(val / 1_000_000).isoformat()
    return val


@dataclass
class Operation:
    """SQL operation."""

    type: OperationType
    model: Type[Model]
    data: dict

    def as_row_dict(self):
        """Serialize a correctly typed DB row from data."""
        res = {}
        for col in self.model.__mapper__.columns:
            if col.name in self.data:
                res[col.name] = to_db_type(col, self.data[col.name])
        return res

    @property
    def pk_dict(self):
        """Primary keys dict."""
        return {
            col.name: self.data[col.name] for col in self.model.__mapper__.primary_key
        }

    @property
    def pk_clauses(self):
        """Primary keys where clauses."""
        return [col == self.data[col.name] for col in self.model.__mapper__.primary_key]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration record table load module."""

from uuid import UUID

from ....load.postgresql.bulk.generators import TableGenerator
from ....state import STATE
from ...models.communities import RDMParentCommunityMetadata
from ...models.records import RDMVersionState


def _is_valid_uuid(value):
    try:
        UUID(value)
        return True
    except Exception:
        return False


class RDMVersionStateTableGenerator(TableGenerator):
    """RDM version state computed table."""

    def __init__(self):
        """Constructor."""
        super().__init__(tables=[RDMVersionState, RDMParentCommunityMetadata])

    def _generate_rows(self, parent_entry, **kwargs):
        # Version state to be populated in the end from the final state
        yield RDMVersionState(
            latest_index=parent_entry.get("latest_index"),
            parent_id=parent_entry["id"],
            latest_id=parent_entry.get("latest_id"),
            next_draft_id=parent_entry.get("next_draft_id"),
        )
        # Generate parent community rows
        communities = parent_entry.get("communities") or []
        for comm_id in communities:
            if _is_valid_uuid(comm_id):
                yield RDMParentCommunityMetadata(
                    record_id=parent_entry["id"],
                    community_id=comm_id,
                )

    def prepare(self, tmp_dir, entry, stack, output_files, **kwargs):
        """Compute rows."""
        # overwrite to avoid it using the parent class
        pass

    def post_prepare(self, tmp_dir, stack, output_files, **kwargs):
        """Overwrite entries with parent state entries."""
        for entry in STATE.PARENTS.all():
            super().prepare(tmp_dir, entry, stack, output_files, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration user table load module."""

from ...load.ids import generate_pk_for, generate_uuid, pid_pk
from ...load.postgresql.bulk.generators import TableGenerator
from ...state import STATE
from ..models.communities import (
    Community,
    CommunityFile,
    CommunityMember,
    FeaturedCommunity,
)
from ..models.files import FilesBucket, FilesObjectVersion
from ..models.oai import OAISet


def _generate_members_uuids(data):
    for member in data["community_members"]:
        member["id"] = generate_uuid(data)
    return data["community_members"]


def _generate_featured_community_id(data):
    if not data.get("featured_community"):
        return None
    return pid_pk()


class CommunityTableGenerator(TableGenerator):
    """User and related tables load."""

    def __init__(self):
        """Constructor."""
        super().__init__(
            tables=[
                Community,
                CommunityMember,
                FeaturedCommunity,
                OAISet,
                FilesBucket,
                FilesObjectVersion,
                CommunityFile,
            ],
            pks=[
                ("community.id", generate_uuid),
                ("oai_set.id", generate_pk_for(OAISet)),
                ("community_members", _generate_members_uuids),
                ("featured_community.id", _generate_featured_community_id),
                ("community_files.file_object.version_id", generate_uuid),
                ("community_files.bucket.id", generate_uuid),
            ],
        )

    def _generate_rows(self, data, **kwargs):
        community = data["community"]
        community_id = community["id"]
        community_slug = community["slug"]
        community_oai_set = data["oai_set"]
        community_files = data["community_files"]
        bucket = community_files["bucket"]

        community["bucket_id"] = bucket["id"]
        yield FilesBucket(**bucket)
        yield Community(**community)

        community_oai_set["search_pattern"] = f"parent.communities.ids:{community_id}"
        community_oai_set["system_created"] = True
        yield OAISet(**community_oai_set)

        community_members = data["community_members"]
        for member in community_members:
            member["community_id"] = community_id
            yield CommunityMember(**member)
        community_owner = next(
            (
                m["user_id"]
                for m in community_members
                if m.get("role") == "owner" and m.get("user_id")
            ),
            None,
        )

        featured_community = data["featured_community"]
        if featured_community.get("id"):
            featured_community["community_id"] = community_id
            yield FeaturedCommunity(**featured_community)

        community_file = community_files.get("file")
        file_object = community_files.get("file_object")
        # Not every community has a logo
        if community_file:
            file_object["bucket_id"] = bucket["id"]
            file_object["file_id"] = community_file["id"]
            yield FilesObjectVersion(**file_object)

            community_file["record_id"] = community_id
            community_file["object_version_id"] = file_object["version_id"]
            yield CommunityFile(**community_file)

        if not STATE.COMMUNITIES.get(community_slug):
            STATE.COMMUNITIES.add(
                community_slug,
                {
                    "id": community_id,
                    "bucket_id": bucket["id"],
                    "owner_id": community_owner,
                    "oai_set_id": community_oai_set["id"],
                    "community_file_id": (community_file or {}).get("id"),
                    "logo_object_version_id": (file_object or {}).get("version_id"),
                },
            )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration drafts row load module."""

from dataclasses import dataclass
from datetime import datetime
from typing import Optional
from uuid import uuid4

import sqlalchemy as sa

from ....actions import LoadAction, LoadData
from ....load.postgresql.transactions.operations import Operation, OperationType
from ...models.communities import Community, CommunityMember
from ...models.files import FilesBucket, FilesObjectVersion
from ...models.pids import PersistentIdentifier
from ...models.records import (
    RDMDraftFile,
    RDMDraftMetadata,
    RDMParentMetadata,
    RDMRecordFile,
    RDMRecordMetadata,
    RDMVersionState,
)


def _get_community_id(session, slug):
    return session.execute(
        sa.select(Community.id).where(Community.slug == slug)
    ).one_or_none()


def _set_permission_flags(session, parent, draft_or_record):
    comm_ids = set(parent.get("communities", {}).get("ids", []))
    if comm_ids:
        permission_flags = {}
        owner_id = parent["json"].get("access", {}).get("owned_by", {}).get("user")
        owner_comm_ids = session.scalars(
            sa.select(Community.id)
            .join(CommunityMember, Community.id == CommunityMember.community_id)
            .where(
                CommunityMember.role == "owner",
                CommunityMember.user_id == owner_id,
            )
        ).all()
        has_only_managed_communities = comm_ids <= set(owner_comm_ids)
        if not has_only_managed_communities:
            permission_flags["can_community_manage_record"] = False
            draft_files_access = draft_or_record["json"].get("access", {}).get("files")
            if draft_files_access != "public":
                permission_flags["can_community_read_files"] = False
        if permission_flags:
            parent["json"]["permission_flags"] = permission_flags


def resolve_communities(session, communities):
    default_slug = communities.get("default")
    if default_slug:
        community_id = _get_community_id(session, default_slug)
        if community_id:
            communities["default"] = community_id
        else:
            communities.pop("default", None)
            communities["ids"].remove(default_slug)

    communities_slugs = communities.get("ids", [])
    _ids = []
    for slug in communities_slugs:
        community_id = _get_community_id(session, slug)
        if community_id:
            _ids.append(community_id)

    communities["ids"] = _ids


def format_pid_ref(pid):
    return {
        "pk": pid.id,
        "obj_type": pid.object_type,
        "pid_type": pid.pid_type,
        "status": pid.status,
    }


def get_published_record(session, recid) -> Optional[RDMRecordMetadata]:
    return session.scalars(
        sa.select(RDMRecordMetadata)
        .join(
            PersistentIdentifier,
            RDMRecordMetadata.id == PersistentIdentifier.object_uuid,
        )
        .where(
            PersistentIdentifier.pid_type == "recid",
            PersistentIdentifier.object_type == "rec",
            PersistentIdentifier.pid_value == recid,
        )
    ).one_or_none()


def get_pid(session, pid_type, pid_value) -> Optional[PersistentIdentifier]:
    return session.scalars(
        sa.select(PersistentIdentifier).where(
            PersistentIdentifier.pid_type == pid_type,
            PersistentIdentifier.pid_value == pid_value,
        )
    ).one_or_none()


def delete_ov(session, bucket_id, key):
    obj = session.scalar(
        sa.select(FilesObjectVersion).where(
            FilesObjectVersion.bucket_id == bucket_id,
            FilesObjectVersion.key == key,
            FilesObjectVersion.is_head.is_(True),
            FilesObjectVersion.file_id.isnot(None),
        )
    )
    if obj:
        yield Operation(
            OperationType.UPDATE,
            FilesObjectVersion,
            {"version_id": obj.version_id, "is_head": False},
        )
        yield Operation(
            OperationType.INSERT,
            FilesObjectVersion,
            {
                "version_id": uuid4(),
                "is_head": True,
            },
        )


def copy_ov(session, src, bucket_id, version_id=None, created=None, updated=None):
    latest_obj = session.scalar(
        sa.select(FilesObjectVersion).where(
            FilesObjectVersion.bucket_id == bucket_id,
            FilesObjectVersion.key == src.key,
            FilesObjectVersion.is_head.is_(True),
        )
    )
    if latest_obj is not None:
        yield Operation(
            OperationType.UPDATE,
            FilesObjectVersion,
            {"version_id": latest_obj.version_id, "is_head": False},
        )

    yield Operation(
        OperationType.INSERT,
        FilesObjectVersion,
        {
            "version_id": version_id or uuid4(),
            "is_head": True,
            "bucket_id": bucket_id,
            "key": src.key,
            "file_id": src.file_id,
            "created": created or datetime.utcnow(),
            "updated": updated or datetime.utcnow(),
        },
    )


def synchronize_files(session, draft, draft_bucket, record, record_bucket):
    draft_ovs = session.scalars(
        sa.select(FilesObjectVersion)
        .join(
            FilesObjectVersion,
            RDMDraftFile.object_version_id == FilesObjectVersion.version_id,
        )
        .where(
            RDMDraftFile.record_id == draft["id"],
            FilesObjectVersion.is_head.is_(True),
        )
    ).all()

    record_ovs = session.scalars(
        sa.select(FilesObjectVersion)
        .join(
            FilesObjectVersion,
            RDMRecordFile.object_version_id == FilesObjectVersion.version_id,
        )
        .where(
            RDMRecordFile.record_id == record["id"],
            FilesObjectVersion.is_head.is_(True),
        )
    ).all()

    changed_ovs = []
    draft_dict_ovs = {}
    for ov in draft_ovs:
        if ov.key not in draft_dict_ovs:
            draft_dict_ovs[ov.key] = ov

    record_dict_ovs = {}
    for ov in record_ovs:
        if ov.key not in record_dict_ovs:
            record_dict_ovs[ov.key] = ov

    for key, ov in draft_dict_ovs.items():
        key_in_dest = key in record_dict_ovs
        if ov.file_id is None:
            if key_in_dest and record_dict_ovs[key].file_id is not None:
                yield from delete_ov(session, record_bucket["id"], key)
                changed_ovs.append(("delete", key))
        else:
            if not key_in_dest:
                new_ov_id = None
                for op in copy_ov(session, ov, record_bucket["id"]):
                    new_ov_id = op.data["version_id"]
                    yield op
                if new_ov_id:
                    changed_ovs.append(("add", new_ov_id))
            else:
                record_ov = record_dict_ovs[key]
                file_in_record_differs = ov.file_id != record_ov.file_id
                ov_deleted_in_dest = record_ov.file_id is None
                if file_in_record_differs or ov_deleted_in_dest:
                    new_ov_id = None
                    for op in copy_ov(session, ov, record_bucket["id"]):
                        new_ov_id = op.data["version_id"]
                        yield op
                    if new_ov_id:
                        if ov_deleted_in_dest:
                            changed_ovs.append(("add", new_ov_id))
                        else:
                            changed_ovs.append(("update", new_ov_id))

    for key, ov in record_dict_ovs.items():
        if key not in draft_dict_ovs:
            yield from delete_ov(session, record_bucket["id"], key)
            changed_ovs.append(("delete", key))
    # TODO: Process `changed_ovs` to apply changes to RDMFileRecord


def resolve_draft_pids(session, draft, parent=None):
    """Enforce record's pids to draft."""
    if not draft:
        return

    pid = draft["json"]["id"]
    forked_published = get_published_record(session, pid)
    if forked_published:
        pids = draft["json"]["pids"]
        has_draft_external_doi = pids.get("doi", {}).get("provider") == "external"
        if has_draft_external_doi:
            # keep the draft external value as it might be there for
            # updating the existing value. Update the draft only with `oai`
            pids["oai"] = forked_published.json["pids"]["oai"]
        else:
            # enfore published record pids to draft
            draft["json"]["pids"] = forked_published.json["pids"]

        if parent:
            parent_model = session.scalars(
                sa.select(RDMParentMetadata).where(
                    RDMParentMetadata.id == forked_published.parent_id
                )
            ).one()
            parent["json"]["pids"] = parent_model.json["pids"]


@dataclass
class RDMDraftCreateData(LoadData):
    """Draft create action data."""

    pid: dict
    draft: dict
    draft_bucket: dict
    parent_pid: dict
    parent: dict


class DraftCreateAction(LoadAction):
    """RDM draft creation."""

    name = "create-draft"
    data_cls = RDMDraftCreateData
    data: RDMDraftCreateData

    def _generate_rows(self, session, **kwargs):
        """Generates rows for a new draft."""
        draft = self.data.draft
        draft["id"] = str(uuid4())
        parent = self.data.parent
        parent.setdefault("id", str(uuid4()))
        pid = self.data.pid

        published_record = get_published_record(session, draft["json"]["id"])
        version_state = session.scalars(
            sa.select(RDMVersionState).where(RDMVersionState.parent_id == parent["id"])
        ).one_or_none()

        is_first_publish_draft = not published_record and not version_state
        is_new_version_draft = version_state and not published_record

        if is_first_publish_draft or is_new_version_draft:
            # Recid PID
            pid.update(
                {
                    "pid_type": "recid",
                    "object_type": "rec",
                    "object_uuid": draft["id"],
                    "status": "R",
                }
            )
            yield Operation(OperationType.INSERT, PersistentIdentifier, pid)
            pid_model = get_pid(session, "recid", pid["pid_value"])
            assert pid_model
            self.data.draft["json"]["pid"] = format_pid_ref(pid_model)
            # Bucket
            yield Operation(OperationType.INSERT, FilesBucket, self.data.draft_bucket)
        else:
            raise Exception("Draft PID creation in invalid state.")

        # Parent
        #  A) draft of a published record, parent id = parent id of published
        #  B) new version, parent id = parent id of the previous version
        #  C) draft of a new record, parent id = given by pk func
        if is_first_publish_draft:
            parent_pid = self.data.parent_pid
            parent_pid.update(
                {
                    "pid_type": "recid",
                    "object_type": "rec",
                    "object_uuid": parent["id"],
                    "status": "R",
                }
            )
            yield Operation(OperationType.INSERT, PersistentIdentifier, parent_pid)
            parent_pid_model = get_pid(session, "recid", parent_pid["pid_value"])
            assert parent_pid_model
            self.data.parent["json"]["pid"] = format_pid_ref(parent_pid_model)
            yield Operation(
                OperationType.INSERT,
                RDMParentMetadata,
                dict(
                    id=parent["id"],
                    json=parent["json"],
                    created=parent["created"],
                    updated=parent["updated"],
                    version_id=parent["version_id"],
                ),
            )
        else:
            raise Exception("Draft parent creation in invalid state.")
        draft["parent_id"] = parent["id"]

        if is_first_publish_draft or is_new_version_draft:
            if is_first_publish_draft:
                draft_index = 1
            else:
                assert version_state
                draft_index = version_state.latest_index + 1
            yield Operation(
                OperationType.INSERT,
                RDMDraftMetadata,
                dict(
                    id=draft["id"],
                    json=draft["json"],
                    created=draft["created"],
                    updated=draft["updated"],
                    version_id=draft["version_id"],
                    index=draft_index,
                    bucket_id=draft["bucket_id"],
                    parent_id=parent["id"],
                    expires_at=draft["expires_at"],
                    fork_version_id=None,
                ),
            )

        if is_first_publish_draft:
            yield Operation(
                OperationType.INSERT,
                RDMVersionState,
                dict(
                    latest_index=None,
                    parent_id=parent["id"],
                    latest_id=None,
                    next_draft_id=draft["id"],
                ),
            )
        elif is_new_version_draft:
            yield Operation(
                OperationType.UPDATE,
                RDMVersionState,
                dict(
                    parent_id=parent["id"],
                    next_draft_id=draft["id"],
                ),
            )

    def _resolve_references(self, session, **kwargs):
        """Resolve references e.g communities slug names."""
        parent = self.data.parent
        parent_pid = get_pid(session, "recid", parent["json"]["id"])
        if parent_pid:
            parent["id"] = str(parent_pid.object_uuid)

        # resolve parent communities slug
        communities = parent["json"].get("communities")
        if communities:
            resolve_communities(session, communities)
        resolve_draft_pids(session, self.data.draft)
        _set_permission_flags(session, parent, self.data.draft)


@dataclass
class RDMDraftEditData(LoadData):
    """Draft edit action data."""

    draft: dict
    parent: dict
    bucket: Optional[dict]


class DraftEditAction(LoadAction):
    """RDM draft edit/update."""

    name = "edit-draft"
    data_cls = RDMDraftEditData
    data: RDMDraftEditData

    def _generate_rows(self, session, **kwargs):
        """Generates rows for a new draft."""
        draft = self.data.draft

        fork_version_id = session.scalar(
            sa.select(RDMRecordMetadata.version_id).where(
                RDMDraftMetadata.id == draft["id"]
            )
        )
        yield Operation(
            OperationType.UPDATE,
            RDMDraftMetadata,
            {
                "id": draft["id"],
                "json": draft["json"],
                "updated": draft["updated"],
                "fork_version_id": fork_version_id,
                RDMDraftMetadata.version_id: RDMDraftMetadata.version_id + 1,
            },
        )
        if self.data.bucket:
            yield Operation(
                OperationType.UPDATE,
                FilesBucket,
                {
                    "id": self.data.bucket["id"],
                    "size": self.data.bucket["size"],
                    "locked": self.data.bucket["locked"],
                    "updated": self.data.bucket["updated"],
                },
            )

    def _resolve_references(self, session, **kwargs):
        """Resolve references e.g communities slug names."""
        draft = self.data.draft
        draft_pid = get_pid(session, "recid", draft["json"]["id"])
        assert draft_pid
        draft["id"] = str(draft_pid.object_uuid)
        draft["index"] = session.scalar(
            sa.select(RDMDraftMetadata.index).where(RDMDraftMetadata.id == draft["id"])
        )
        draft["json"]["pid"] = format_pid_ref(draft_pid)

        parent = self.data.parent
        parent_pid = get_pid(session, "recid", parent["json"]["id"])
        assert parent_pid
        parent["id"] = str(parent_pid.object_uuid)
        parent["json"]["pid"] = format_pid_ref(parent_pid)

        # resolve parent communities slug
        communities = parent.get("json", {}).get("communities")
        if communities:
            resolve_communities(session, communities)

        resolve_draft_pids(session, self.data.draft, parent=parent)
        _set_permission_flags(session, parent, self.data.draft)


@dataclass
class RDMDraftPublishNewData(LoadData):
    """Draft new publish action data."""

    # Files
    draft_bucket: dict
    record_bucket: dict
    record_object_versions: list[dict]

    # Records
    parent: dict
    draft: dict
    record: dict

    # PIDs
    pid: dict
    oai_pid: dict
    doi: dict
    parent_pid: Optional[dict] = None
    parent_doi: Optional[dict] = None

    # Communities?


class DraftPublishNewAction(LoadAction):
    """RDM new publish action."""

    name = "publish-new-draft"
    data_cls = RDMDraftPublishNewData
    data: RDMDraftPublishNewData

    def _generate_rows(self, session, **kwargs):
        """Generates rows for a new draft."""
        is_first_publish = self.data.parent_pid is not None
        is_local_doi = self.data.parent_doi is not None

        draft = self.data.draft
        record = self.data.record
        parent = self.data.parent
        #
        # Files rows
        #
        draft_bucket = self.data.draft_bucket
        record_bucket = self.data.record_bucket

        record_bucket["locked"] = True
        yield Operation(OperationType.INSERT, FilesBucket, record_bucket)
        yield Operation(
            OperationType.UPDATE,
            FilesBucket,
            {
                "id": draft_bucket["id"],
                "updated": draft_bucket["updated"],
                "locked": True,
            },
        )

        record_object_versions = self.data.record_object_versions
        for record_ov in record_object_versions:
            yield Operation(OperationType.INSERT, FilesObjectVersion, record_ov)
            yield Operation(
                OperationType.INSERT,
                RDMRecordFile,
                {
                    "id": str(uuid4()),
                    "created": record_ov["created"],
                    "updated": record_ov["updated"],
                    "json": {},
                    "version_id": 1,
                    "key": record_ov["key"],
                    "record_id": record["id"],
                    "object_version_id": record_ov["version_id"],
                },
            )

        #
        # PID rows
        #
        pid = self.data.pid
        pid_model = get_pid(session, "recid", pid["pid_value"])
        assert pid_model
        yield Operation(
            OperationType.UPDATE,
            PersistentIdentifier,
            {
                "id": pid_model.id,
                "status": "R",
                "updated": pid["updated"],
            },
        )

        doi = self.data.doi
        doi["object_uuid"] = record["id"]
        doi["status"] = "R"
        yield Operation(OperationType.INSERT, PersistentIdentifier, doi)

        oai_pid = self.data.oai_pid
        oai_pid["status"] = "R"
        oai_pid["object_uuid"] = record["id"]
        yield Operation(OperationType.INSERT, PersistentIdentifier, oai_pid)

        if is_first_publish:
            parent_pid = self.data.parent_pid
            assert parent_pid
            parent_pid_model = get_pid(session, "recid", parent_pid["pid_value"])
            assert parent_pid_model
            yield Operation(
                OperationType.UPDATE,
                PersistentIdentifier,
                {
                    "id": parent_pid_model.id,
                    "status": "R",
                    "updated": parent_pid["updated"],
                },
            )
            if is_local_doi:
                parent_doi = self.data.parent_doi
                assert parent_doi
                parent_doi["object_uuid"] = parent["id"]
                parent_doi["status"] = "R"
                yield Operation(OperationType.INSERT, PersistentIdentifier, parent_doi)

        #
        # Draft rows
        #
        # Soft-delete draft
        yield Operation(
            OperationType.UPDATE,
            RDMDraftMetadata,
            {
                "id": draft["id"],
                "json": None,
                "updated": draft["updated"],
                RDMDraftMetadata.version_id: RDMDraftMetadata.version_id + 1,
                "fork_version_id": None,
            },
        )

        # Create record
        yield Operation(
            OperationType.INSERT,
            RDMRecordMetadata,
            dict(
                id=record["id"],
                json=record["json"],
                created=record["updated"],
                updated=record["updated"],
                version_id=1,
                index=draft["index"],  # take from drat
                bucket_id=record_bucket["id"],
                parent_id=parent["id"],
                deletion_status="P",
            ),
        )
        # Update parent
        yield Operation(
            OperationType.UPDATE,
            RDMParentMetadata,
            {
                "id": parent["id"],
                "json": parent["json"],
                "updated": parent["updated"],
                RDMParentMetadata.version_id: RDMParentMetadata.version_id + 1,
            },
        )
        yield Operation(
            OperationType.UPDATE,
            RDMVersionState,
            dict(
                parent_id=parent["id"],
                latest_id=record["id"],
                latest_index=draft["index"],
                next_draft_id=None,
            ),
        )

    def _resolve_references(self, session, **kwargs):
        """Resolve references e.g communities slug names."""
        draft = self.data.draft
        draft_pid = get_pid(session, "recid", draft["json"]["id"])
        assert draft_pid
        draft["id"] = str(draft_pid.object_uuid)
        draft["index"] = session.scalar(
            sa.select(RDMDraftMetadata.index).where(RDMDraftMetadata.id == draft["id"])
        )
        draft["json"]["pid"] = format_pid_ref(draft_pid)
        # In RDM drafts and records share their PK
        self.data.record["id"] = draft["id"]
        self.data.record["json"]["pid"] = draft["json"]["pid"]

        parent = self.data.parent
        parent_pid = get_pid(session, "recid", parent["json"]["id"])
        assert parent_pid
        parent["id"] = str(parent_pid.object_uuid)
        parent["json"]["pid"] = format_pid_ref(parent_pid)

        # resolve parent communities slug
        communities = parent["json"].get("communities")
        if communities:
            resolve_communities(session, communities)
        resolve_draft_pids(session, self.data.draft, parent=parent)
        _set_permission_flags(session, parent, self.data.draft)


@dataclass
class RDMDraftPublishEditData(LoadData):
    """Draft edit publish action data."""

    # Files
    draft_bucket: Optional[dict]
    record_bucket: Optional[dict]
    record_object_versions: Optional[dict]

    # Records
    parent: dict
    draft: dict
    record: dict

    # PIDs
    old_external_doi: Optional[dict]
    new_external_doi: Optional[dict]

    # Communities?


class DraftPublishEditAction(LoadAction):
    """RDM edit publish creation."""

    name = "publish-edit-draft"
    data_cls = RDMDraftPublishEditData
    data: RDMDraftPublishEditData

    def _generate_rows(self, session, **kwargs):
        """Generates rows for a new draft."""
        is_external_doi = self.data.old_external_doi or self.data.old_external_doi

        draft = self.data.draft
        record = self.data.record
        parent = self.data.parent

        #
        # Files rows
        #
        if is_external_doi and self.data.record_object_versions:
            # TODO: Synchronize files
            yield from synchronize_files(
                session,
                draft,
                self.data.draft_bucket,
                record,
                self.data.record_bucket,
            )
        #
        # PID rows
        #
        external_doi_changed = self.data.old_external_doi and self.data.new_external_doi
        if external_doi_changed:
            old_doi = self.data.old_external_doi
            assert old_doi
            old_doi_model = get_pid(session, "doi", old_doi["pid_value"])
            assert old_doi_model
            yield Operation(
                OperationType.DELETE, PersistentIdentifier, {"id": old_doi_model.id}
            )
            new_doi = self.data.new_external_doi
            assert new_doi
            new_doi["object_uuid"] = record["id"]
            yield Operation(OperationType.INSERT, PersistentIdentifier, new_doi)

        #
        # Draft rows
        #
        # Soft-delete draft
        yield Operation(
            OperationType.UPDATE,
            RDMDraftMetadata,
            dict(
                id=draft["id"],
                json=None,
                updated=draft["updated"],
                version_id=draft["version_id"],
                fork_version_id=None,
            ),
        )

        # Create record
        yield Operation(
            OperationType.UPDATE,
            RDMRecordMetadata,
            {
                "id": record["id"],
                "json": record["json"],
                "updated": record["updated"],
                RDMRecordMetadata.version_id: RDMRecordMetadata.version_id + 1,
            },
        )

        # Update parent
        yield Operation(
            OperationType.UPDATE,
            RDMParentMetadata,
            {
                "id": parent["id"],
                "json": parent["json"],
                RDMParentMetadata.version_id: RDMParentMetadata.version_id + 1,
                "updated": parent["updated"],
            },
        )

    def _resolve_references(self, session, **kwargs):
        """Resolve references e.g communities slug names."""
        draft = self.data.draft
        draft_pid = get_pid(session, "recid", draft["json"]["id"])
        assert draft_pid
        draft["id"] = str(draft_pid.object_uuid)
        draft["json"]["pid"] = format_pid_ref(draft_pid)
        self.data.record["id"] = draft["id"]
        self.data.record["id"] = draft["id"]
        self.data.record["json"]["pid"] = draft["json"]["pid"]

        parent = self.data.parent
        parent_pid = get_pid(session, "recid", parent["json"]["id"])
        assert parent_pid
        parent["id"] = str(parent_pid.object_uuid)
        parent["json"]["pid"] = format_pid_ref(parent_pid)

        # resolve parent communities slug
        communities = parent["json"].get("communities")
        if communities:
            resolve_communities(session, communities)
        resolve_draft_pids(session, self.data.draft, parent=parent)
        _set_permission_flags(session, parent, self.data.draft)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""GitHub actions tests."""

import pytest

from invenio_rdm_migrator.load.postgresql.transactions.operations import OperationType
from invenio_rdm_migrator.streams.actions.load import (
    HookEventCreateAction,
    HookEventUpdateAction,
    ReleaseReceiveAction,
    ReleaseUpdateAction,
    RepoCreateAction,
    RepoUpdateAction,
)
from invenio_rdm_migrator.streams.models.github import Release, Repository, WebhookEvent
from invenio_rdm_migrator.streams.models.oauth import ServerToken

#
# Hooks
#


@pytest.fixture(scope="function")
def gh_repo_data():
    """GitHub repository data."""
    return {
        "created": "2022-01-01T00:00:00",
        "updated": "2022-01-01T00:00:00",
        "id": "0d1b629d-7992-4650-b0b0-8908a0322bca",
        "github_id": 427018972,
        "name": "ppanero/zenodo-release-test",
        # the following None means the hook was removed, but it does not affect the tests
        "user_id": None,
        "hook": None,
        # note: ping was removed
    }


@pytest.fixture(scope="function")
def event_data():
    """Webhook event data."""
    return {
        "created": "2022-01-01T00:00:00",
        "updated": "2022-01-01T00:00:00",
        "id": "189d88dd-22d9-40d1-b3af-9da4b2bc4870",
        "receiver_id": "github",
        "user_id": 86490,
        "payload": '{"action": "published", "release": {"body": "Zenodo testing migration", "assets_url": "https://api.github.com/repos/ppanero/zenodo-release-test/releases/121854239/assets", "name": "more and more", "published_at": "2023-09-20T11:52:47Z", "author": {"following_url": "https://api.github.com/users/ppanero/following{/other_user}", "events_url": "https://api.github.com/users/ppanero/events{/privacy}", "avatar_url": "https://avatars.githubusercontent.com/u/6756943?v=4", "url": "https://api.github.com/users/ppanero", "gists_url": "https://api.github.com/users/ppanero/gists{/gist_id}", "html_url": "https://github.com/ppanero", "subscriptions_url": "https://api.github.com/users/ppanero/subscriptions", "node_id": "MDQ6VXNlcjY3NTY5NDM=", "repos_url": "https://api.github.com/users/ppanero/repos", "received_events_url": "https://api.github.com/users/ppanero/received_events", "gravatar_id": "", "starred_url": "https://api.github.com/users/ppanero/starred{/owner}{/repo}", "site_admin": false, "login": "ppanero", "type": "User", "id": 6756943, "followers_url": "https://api.github.com/users/ppanero/followers", "organizations_url": "https://api.github.com/users/ppanero/orgs"}, "url": "https://api.github.com/repos/ppanero/zenodo-release-test/releases/121854239", "created_at": "2023-09-20T11:50:17Z", "target_commitish": "main", "tarball_url": "https://api.github.com/repos/ppanero/zenodo-release-test/tarball/v4", "html_url": "https://github.com/ppanero/zenodo-release-test/releases/tag/v4", "zipball_url": "https://api.github.com/repos/ppanero/zenodo-release-test/zipball/v4", "tag_name": "v4", "node_id": "RE_kwDOGXPK3M4HQ1kf", "draft": false, "prerelease": false, "upload_url": "https://uploads.github.com/repos/ppanero/zenodo-release-test/releases/121854239/assets{?name,label}", "id": 121854239, "assets": []}, "sender": {"following_url": "https://api.github.com/users/ppanero/following{/other_user}", "events_url": "https://api.github.com/users/ppanero/events{/privacy}", "avatar_url": "https://avatars.githubusercontent.com/u/6756943?v=4", "url": "https://api.github.com/users/ppanero", "gists_url": "https://api.github.com/users/ppanero/gists{/gist_id}", "html_url": "https://github.com/ppanero", "subscriptions_url": "https://api.github.com/users/ppanero/subscriptions", "node_id": "MDQ6VXNlcjY3NTY5NDM=", "repos_url": "https://api.github.com/users/ppanero/repos", "received_events_url": "https://api.github.com/users/ppanero/received_events", "gravatar_id": "", "starred_url": "https://api.github.com/users/ppanero/starred{/owner}{/repo}", "site_admin": false, "login": "ppanero", "type": "User", "id": 6756943, "followers_url": "https://api.github.com/users/ppanero/followers", "organizations_url": "https://api.github.com/users/ppanero/orgs"}, "repository": {"issues_url": "https://api.github.com/repos/ppanero/zenodo-release-test/issues{/number}", "deployments_url": "https://api.github.com/repos/ppanero/zenodo-release-test/deployments", "stargazers_count": 0, "forks_url": "https://api.github.com/repos/ppanero/zenodo-release-test/forks", "mirror_url": null, "allow_forking": true, "subscription_url": "https://api.github.com/repos/ppanero/zenodo-release-test/subscription", "topics": [], "notifications_url": "https://api.github.com/repos/ppanero/zenodo-release-test/notifications{?since,all,participating}", "collaborators_url": "https://api.github.com/repos/ppanero/zenodo-release-test/collaborators{/collaborator}", "updated_at": "2021-12-02T13:03:26Z", "private": false, "pulls_url": "https://api.github.com/repos/ppanero/zenodo-release-test/pulls{/number}", "disabled": false, "issue_comment_url": "https://api.github.com/repos/ppanero/zenodo-release-test/issues/comments{/number}", "labels_url": "https://api.github.com/repos/ppanero/zenodo-release-test/labels{/name}", "has_wiki": true, "full_name": "ppanero/zenodo-release-test", "owner": {"following_url": "https://api.github.com/users/ppanero/following{/other_user}", "events_url": "https://api.github.com/users/ppanero/events{/privacy}", "avatar_url": "https://avatars.githubusercontent.com/u/6756943?v=4", "url": "https://api.github.com/users/ppanero", "gists_url": "https://api.github.com/users/ppanero/gists{/gist_id}", "html_url": "https://github.com/ppanero", "subscriptions_url": "https://api.github.com/users/ppanero/subscriptions", "node_id": "MDQ6VXNlcjY3NTY5NDM=", "repos_url": "https://api.github.com/users/ppanero/repos", "received_events_url": "https://api.github.com/users/ppanero/received_events", "gravatar_id": "", "starred_url": "https://api.github.com/users/ppanero/starred{/owner}{/repo}", "site_admin": false, "login": "ppanero", "type": "User", "id": 6756943, "followers_url": "https://api.github.com/users/ppanero/followers", "organizations_url": "https://api.github.com/users/ppanero/orgs"}, "statuses_url": "https://api.github.com/repos/ppanero/zenodo-release-test/statuses/{sha}", "id": 427018972, "keys_url": "https://api.github.com/repos/ppanero/zenodo-release-test/keys{/key_id}", "description": null, "tags_url": "https://api.github.com/repos/ppanero/zenodo-release-test/tags", "archived": false, "downloads_url": "https://api.github.com/repos/ppanero/zenodo-release-test/downloads", "assignees_url": "https://api.github.com/repos/ppanero/zenodo-release-test/assignees{/user}", "watchers": 0, "contents_url": "https://api.github.com/repos/ppanero/zenodo-release-test/contents/{+path}", "has_pages": false, "git_refs_url": "https://api.github.com/repos/ppanero/zenodo-release-test/git/refs{/sha}", "has_discussions": false, "has_projects": true, "clone_url": "https://github.com/ppanero/zenodo-release-test.git", "watchers_count": 0, "git_tags_url": "https://api.github.com/repos/ppanero/zenodo-release-test/git/tags{/sha}", "milestones_url": "https://api.github.com/repos/ppanero/zenodo-release-test/milestones{/number}", "languages_url": "https://api.github.com/repos/ppanero/zenodo-release-test/languages", "size": 8, "homepage": null, "fork": false, "commits_url": "https://api.github.com/repos/ppanero/zenodo-release-test/commits{/sha}", "releases_url": "https://api.github.com/repos/ppanero/zenodo-release-test/releases{/id}", "issue_events_url": "https://api.github.com/repos/ppanero/zenodo-release-test/issues/events{/number}", "archive_url": "https://api.github.com/repos/ppanero/zenodo-release-test/{archive_format}{/ref}", "comments_url": "https://api.github.com/repos/ppanero/zenodo-release-test/comments{/number}", "events_url": "https://api.github.com/repos/ppanero/zenodo-release-test/events", "contributors_url": "https://api.github.com/repos/ppanero/zenodo-release-test/contributors", "html_url": "https://github.com/ppanero/zenodo-release-test", "visibility": "public", "forks": 0, "compare_url": "https://api.github.com/repos/ppanero/zenodo-release-test/compare/{base}...{head}", "open_issues": 0, "node_id": "R_kgDOGXPK3A", "git_url": "git://github.com/ppanero/zenodo-release-test.git", "svn_url": "https://github.com/ppanero/zenodo-release-test", "merges_url": "https://api.github.com/repos/ppanero/zenodo-release-test/merges", "has_issues": true, "ssh_url": "git@github.com:ppanero/zenodo-release-test.git", "blobs_url": "https://api.github.com/repos/ppanero/zenodo-release-test/git/blobs{/sha}", "git_commits_url": "https://api.github.com/repos/ppanero/zenodo-release-test/git/commits{/sha}", "hooks_url": "https://api.github.com/repos/ppanero/zenodo-release-test/hooks", "has_downloads": true, "license": {"spdx_id": "MPL-2.0", "url": "https://api.github.com/licenses/mpl-2.0", "node_id": "MDc6TGljZW5zZTE0", "name": "Mozilla Public License 2.0", "key": "mpl-2.0"}, "name": "zenodo-release-test", "language": null, "url": "https://api.github.com/repos/ppanero/zenodo-release-test", "created_at": "2021-11-11T13:53:02Z", "open_issues_count": 0, "is_template": false, "pushed_at": "2023-09-20T11:50:38Z", "web_commit_signoff_required": false, "forks_count": 0, "default_branch": "main", "teams_url": "https://api.github.com/repos/ppanero/zenodo-release-test/teams", "trees_url": "https://api.github.com/repos/ppanero/zenodo-release-test/git/trees{/sha}", "branches_url": "https://api.github.com/repos/ppanero/zenodo-release-test/branches{/branch}", "subscribers_url": "https://api.github.com/repos/ppanero/zenodo-release-test/subscribers", "stargazers_url": "https://api.github.com/repos/ppanero/zenodo-release-test/stargazers"}}',
        "payload_headers": None,
        "response": '{"status": 409, "message": "The release has already been received."}',
        "response_headers": None,
        "response_code": 409,
    }


def test_github_repo_update(session, gh_repo_data):
    data = dict(gh_repository=gh_repo_data)
    action = RepoUpdateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == Repository


def test_github_hook_event_create_wo_token(session, event_data):
    data = dict(webhook_event=event_data)
    action = HookEventCreateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.INSERT
    assert rows[0].model == WebhookEvent


def test_github_hook_event_create_w_token(session, event_data):
    data = dict(webhook_event=event_data)
    action = HookEventCreateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.INSERT
    assert rows[0].model == WebhookEvent


def test_github_hook_event_update(session, event_data):
    data = dict(webhook_event=event_data)
    action = HookEventUpdateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == WebhookEvent


#
# Releases
#


@pytest.fixture(scope="function")
def gh_release_data():
    """GitHub repository data."""
    return {
        "created": "2022-01-01T00:00:00",
        "updated": "2022-01-01T00:00:00",
        "id": "c9fc85cd-e8ec-4ba0-9a13-75a590f3fd15",
        "release_id": 121854239,
        "tag": "v4",
        "errors": None,
        "repository_id": "0d1b629d-7992-4650-b0b0-8908a0322bca",
        "event_id": "1e596bad-1bb3-4749-8a5e-dd9f1552ebc2",
        "record_id": None,
        "status": "R",
    }


def test_github_release_receive(session, gh_repo_data, gh_release_data):
    data = dict(gh_release=gh_release_data, gh_repository=gh_repo_data)
    action = ReleaseReceiveAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 2
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == Repository
    assert rows[1].type == OperationType.INSERT
    assert rows[1].model == Release


def test_github_release_update(session, gh_release_data):
    data = dict(gh_release=gh_release_data)
    action = ReleaseUpdateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == Release



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration draft table load module."""

from datetime import datetime
from functools import partial

from ....load.ids import generate_recid, generate_uuid
from ....load.postgresql.bulk.generators import TableGenerator
from ....state import STATE
from ...models.pids import PersistentIdentifier
from ...models.records import (
    RDMDraftFile,
    RDMDraftMediaFile,
    RDMDraftMetadata,
    RDMParentMetadata,
)
from .parents import generate_parent_rows
from .references import CommunitiesReferencesMixin, PIDsReferencesMixin
from .utils import InsertRecordFiles


class RDMDraftTableGenerator(
    TableGenerator,
    CommunitiesReferencesMixin,
    PIDsReferencesMixin,
):
    """RDM Record and related tables load."""

    def __init__(self):
        """Constructor."""
        super().__init__(
            tables=[
                RDMDraftMetadata,
                RDMParentMetadata,
                PersistentIdentifier,
            ],
            post_load_hooks=[
                InsertRecordFiles(RDMDraftMetadata, RDMDraftFile),
                InsertRecordFiles(
                    RDMDraftMetadata,
                    RDMDraftMediaFile,
                    bucket_fk="media_bucket_id",
                ),
            ],
            pks=[
                ("draft.id", generate_uuid),
                ("parent.id", generate_uuid),
                ("draft.json.pid", partial(generate_recid, status="N")),
                ("parent.json.pid", generate_recid),
                ("draft.parent_id", lambda d: d["parent"]["id"]),
            ],
        )

    def _generate_rows(self, data, **kwargs):
        """Generates rows for a record."""
        now = datetime.utcnow().isoformat()
        parent = data["parent"]
        draft = data.get("draft")
        if not draft:
            return

        recid = draft["json"]["id"]
        forked_published = STATE.RECORDS.get(recid)
        state_parent = STATE.PARENTS.get(parent["json"]["id"])

        if not state_parent:
            STATE.PARENTS.add(
                parent["json"]["id"],  # recid
                {
                    "id": parent["id"],
                    "next_draft_id": draft["id"],
                },
            )
            # drafts have a parent on save
            # on the other hand there is no community parent/request
            yield from generate_parent_rows(parent)
        # if there is a parent but there is no record it means that it is a
        # draft of a new version
        elif state_parent and not forked_published:
            STATE.PARENTS.update(parent["json"]["id"], {"next_draft_id": draft["id"]})

        # if its a draft of a published record, its parent should be parent id
        # if its a new version, its parent should be the one of the previous version
        # otherwise is a new parent (new record, new draft...)
        parent_id = state_parent["id"] if state_parent else draft["parent_id"]
        if forked_published:
            parent_id = forked_published["parent_id"]

        draft_id = forked_published.get("id") or draft["id"]
        # all drafts have a bucket due to the `post_create` hook on the files sysfield
        STATE.BUCKETS.add(draft["bucket_id"], {"draft_id": draft_id})

        yield RDMDraftMetadata(
            id=draft_id,
            json=draft["json"],
            created=draft["created"],
            updated=draft["updated"],
            version_id=draft["version_id"],
            index=forked_published.get("index") or draft["index"],
            bucket_id=draft["bucket_id"],
            media_bucket_id=draft.get("media_bucket_id"),
            parent_id=parent_id,
            expires_at=draft["expires_at"],
            fork_version_id=forked_published.get("fork_version_id")
            or draft["fork_version_id"],
        )

        # if there is a record in the state it means both recid and doi were already
        # processed in the records table generator, a duplicate would violate unique
        # constraints and cause the load to fail.
        if not forked_published:
            # recid
            record_pid = draft["json"]["pid"]
            yield PersistentIdentifier(
                id=record_pid["pk"],
                pid_type=record_pid["pid_type"],  # in drafts are recid
                pid_value=draft["json"]["id"],
                status=record_pid["status"],
                object_type=record_pid["obj_type"],
                object_uuid=draft["id"],
                created=now,
                updated=now,
            )
        # we don't emit doi Persistentidentifier for drafts as either they have already
        # one from records or have an external doi that is registered on publish
        # draft files are a post_hook

    def _resolve_references(self, data, **kwargs):
        """Resolve references e.g communities slug names."""
        # resolve parent communities slug
        draft = data.get("draft")
        if draft:
            parent = data["parent"]
            communities = parent["json"].get("communities")
            if communities:
                self.resolve_communities(communities)
            self.resolve_draft_pids(draft)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration transaction extract classes."""

import copy
from dataclasses import dataclass
from typing import Optional, Sequence, Union

import dictdiffer


@dataclass
class Tx:
    """An extracted DB transaction."""

    id: int
    operations: list[dict]  # TODO: we could more narrowly define it later
    commit_lsn: Optional[int] = None

    def as_ops_tuples(
        self,
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
    ):
        """Return a list of (table, op_type) tuples."""
        res = [(o["source"]["table"], o["op"]) for o in self.operations]
        if include:
            res = [t for t in res if t[0] in include]
        if exclude:
            res = [t for t in res if t[0] not in exclude]
        return res

    def ops_by(
        self,
        table,
        op_types: Optional[Sequence[str]] = ("C", "U"),
        group_id: Optional[Union[Sequence[str], str, bool]] = True,
        group_key: Optional[Union[Sequence[str], str]] = None,
        filter_unchanged=True,
    ):
        """Return rolled-up and/or grouped table operations."""
        result_by_id = {}
        result = {}

        for op in self.operations:
            if op["source"]["table"] == table and op["op"] in op_types:
                if isinstance(group_id, str):
                    group_id = (group_id,)
                if group_id is True or group_id is None:
                    group_id = sorted(op["key"].keys())
                assert isinstance(group_id, (Sequence, str))
                group_key = group_key or group_id
                if isinstance(group_key, str):
                    group_key = (group_key,)

                if filter_unchanged and op["op"] == "U":
                    ignored_keys = set(group_key) | set(group_id)
                    before, after = self.filter_unchanged(op, ignored_keys)
                else:
                    before = op["before"]
                    after = op["after"]
                data = after or before
                data_id = tuple(data[k] for k in group_id)
                # for one-column PKs just use the single value
                if len(data_id) == 1:
                    data_id = data_id[0]

                result_by_id.setdefault(data_id, {})
                result_by_id[data_id].update(data)

                aggr_data = result_by_id[data_id]
                data_key = tuple(aggr_data.get(k) for k in group_key)
                # for one-column keys just use the single value
                if len(data_key) == 1:
                    data_key = data_key[0]
                result[data_key] = aggr_data
        return result

    def filter_ops(self, table, filter: dict):
        """Return rolled-up and/or grouped table operations."""
        return [
            o
            for o in self.operations
            if o["source"]["table"] == table
            and filter.items() <= (o["after"] or o["before"]).items()
        ]

    @staticmethod
    def filter_unchanged(data, ignored_keys=None):
        """Filter out unchanged fields in an transaction operation, keeping PKs."""
        before = copy.deepcopy(data["before"])
        after = copy.deepcopy(data["after"])
        ignored_keys = set(ignored_keys or data["key"].keys())
        diff = dictdiffer.diff(before, after, ignore=ignored_keys)
        changed_keys = {key for diff_op, key, _ in diff if diff_op == "change"}
        for key in (before.keys() | after.keys()) - (changed_keys | ignored_keys):
            before.pop(key)
            after.pop(key)
        return before, after



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration requests table load module."""


from ...load.ids import generate_uuid
from ...load.postgresql.bulk.generators import TableGenerator
from ...state import STATE
from ..models.requests import RequestMetadata


class RequestTableGenerator(TableGenerator):
    """Requests and related tables load."""

    def __init__(self):
        """Constructor."""
        super().__init__(
            tables=[RequestMetadata],
            pks=[("id", generate_uuid)],
        )

    def _generate_rows(self, data, **kwargs):
        """Yield requests metadata."""
        request = data
        yield RequestMetadata(
            id=request["id"],
            json=request["json"],
            created=request["created"],
            updated=request["updated"],
            version_id=request["version_id"],
            number=request["number"],
            expires_at=request["expires_at"],
        )

    def _resolve_references(self, data, **kwargs):
        """Resolve references.

        Translates community slugs to uuids, and record parent pids to uuids. Both values
        are mandatory, therefore they are access as dict.

        :raises: KeyError if the the parent or community are not found in the state.
        """
        # it assumes the data is transformed by an InclusionRequestEntry
        request_slug = data["json"]["receiver"]["community"]
        community = STATE.COMMUNITIES.get(request_slug)

        data["json"]["receiver"]["community"] = community["id"]

    def prepare(self, tmp_dir, entry, stack, output_files, **kwargs):
        """Compute rows."""
        super().prepare(tmp_dir, entry, stack, output_files, create=True, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration files load module."""

from ...load.postgresql.bulk import PostgreSQLCopyLoad
from ...load.postgresql.bulk.generators import ExistingDataTableGenerator
from ..models.files import FilesBucket, FilesInstance, FilesObjectVersion
from .table_generator import FilesTableGenerator


class FilesCopyLoad(PostgreSQLCopyLoad):
    """PostgreSQL files COPY load."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(table_generators=[FilesTableGenerator()], **kwargs)


class ExistingFilesLoad(PostgreSQLCopyLoad):
    """Existing files class for data loading."""

    def __init__(self, db_uri, data_dir, **kwargs):
        """Constructor."""
        super().__init__(
            db_uri=db_uri,
            table_generators=[
                ExistingDataTableGenerator(
                    tables=[FilesInstance, FilesBucket, FilesObjectVersion], pks=[]
                )
            ],
            data_dir=data_dir,
            existing_data=True,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Pytest files configuration."""


import pytest


@pytest.fixture(scope="module")
def transformed_files_entry():
    """Files data as after being transformed.

    These are sample files related information.
    """
    return {
        "bucket": {
            "id": "1",
            "created": "2023-04-19",
            "updated": "2023-04-19",
            "default_location": "1",
            "default_storage_class": "L",
            "size": 1234,
            "quota_size": None,
            "max_file_size": None,
            "locked": False,
            "deleted": False,
        },
        "object_version": {
            "version_id": "2",
            "created": "2023-04-19",
            "updated": "2023-04-19",
            "key": "file.txt",
            "bucket_id": "1",
            "file_id": "3",
            "_mimetype": None,
            "is_head": True,
        },
        "file": {
            "id": "3",
            "created": "2023-04-19",
            "updated": "2023-04-19",
            "uri": "path/to/file",
            "storage_class": "L",
            "size": 1234,
            "checksum": "md5:abcd1234",
            "readable": True,
            "writable": True,
            "last_check_at": None,
            "last_check": False,
        },
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Identity transformation tests."""

from invenio_rdm_migrator.transform import IdentityTransform


def test_identity_transform():
    input = {"implemented": "test", "unimplemented": "test too"}
    expected = {"implemented": "test", "unimplemented": "test too"}

    assert IdentityTransform()._transform(input) == expected



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration parent table load module."""

from datetime import datetime

from ....load.postgresql.transactions.operations import Operation, OperationType
from ...models.pids import PersistentIdentifier
from ...models.records import RDMParentMetadata


def generate_parent_ops(parent, parent_pid):
    """Generates operations for a parent record."""
    now = datetime.utcnow().isoformat()
    # order is important when doing action/streaming migration
    # parent recid
    yield Operation(
        OperationType.INSERT,
        PersistentIdentifier,
        dict(
            id=parent_pid["id"],
            pid_type=parent_pid["pid_type"],
            pid_value=parent_pid["pid_value"],
            status=parent_pid["status"],
            object_type=parent_pid["object_type"],
            object_uuid=parent["id"],
            created=now,
            updated=now,
        ),
    )
    # parent record
    yield Operation(
        OperationType.INSERT,
        RDMParentMetadata,
        dict(
            id=parent["id"],
            json=parent["json"],
            created=parent["created"],
            updated=parent["updated"],
            version_id=parent["version_id"],
        ),
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""File actions tests."""

import pytest

from invenio_rdm_migrator.load.postgresql.transactions.operations import OperationType
from invenio_rdm_migrator.streams.actions.load import (
    OAuthApplicationCreateAction,
    OAuthApplicationDeleteAction,
    OAuthApplicationUpdateAction,
    OAuthGHDisconnectToken,
    OAuthLinkedAccountConnectAction,
    OAuthLinkedAccountDisconnectAction,
    OAuthServerTokenCreateAction,
    OAuthServerTokenDeleteAction,
    OAuthServerTokenUpdateAction,
)
from invenio_rdm_migrator.streams.models.oauth import (
    RemoteAccount,
    RemoteToken,
    ServerClient,
    ServerToken,
)
from invenio_rdm_migrator.streams.models.users import UserIdentity


@pytest.fixture(scope="function")
def oauth_client_data():
    """OAuth2Server client data."""

    return {
        "name": "test-incremental-token",
        "description": "",
        "website": "",
        "user_id": 123456,
        "client_id": "cH4ng3DZPeBjqj7uMB1JWXavhxebu6V0mwMtvMr",
        "client_secret": "cH4ng3D143BM6gKc29VN0rWZPI4wi0gHBcJQYdVNLtibTK0AR1ZWbWT5oYeQ",
        "is_confidential": False,
        "is_internal": True,
        "_redirect_uris": None,
        "_default_scopes": "tokens:generate user:email",
    }


@pytest.fixture(scope="function")
def oauth_token_data():
    """OAuth2Server token data."""

    return {
        "id": 156666,
        "client_id": "cH4ng3DZPeBjqj7uMB1JWXavhxebu6V0mwMtvMr",
        "user_id": 123456,
        "token_type": "bearer",
        "access_token": "cH4ng3DzbXd4QTcrRjFMcTVMRHl3QlY2Rkdib0VwREY4aDhPcHo2dUt2ZnZ3OVVPa1BvRDl0L1NRZmFrdXNIU2hJR2JWc0NHZDZSVEhVT2JQcmdjS1E9PQ==",
        "refresh_token": None,
        "expires": None,
        "_scopes": "tokens:generate user:email",
        "is_personal": True,
        "is_internal": False,
    }


def test_create_oauth_server_token(session, oauth_client_data, oauth_token_data):
    data = dict(
        client=oauth_client_data,
        token=oauth_token_data,
    )
    action = OAuthServerTokenCreateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 2
    assert rows[0].type == OperationType.INSERT
    assert rows[0].model == ServerClient
    assert rows[1].type == OperationType.INSERT
    assert rows[1].model == ServerToken


def test_client_oauth_server_token(session, oauth_client_data, oauth_token_data):
    data = dict(
        client=oauth_client_data,
        token=oauth_token_data,
    )
    action = OAuthServerTokenUpdateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 2
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == ServerClient
    assert rows[1].type == OperationType.UPDATE
    assert rows[1].model == ServerToken


def test_client_oauth_server_token_only_client(session, oauth_client_data):
    data = dict(
        client=oauth_client_data,
    )
    action = OAuthServerTokenUpdateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == ServerClient


def test_client_oauth_server_token_only_server(session, oauth_token_data):
    data = dict(
        token=oauth_token_data,
    )
    action = OAuthServerTokenUpdateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == ServerToken


def test_delete_oauth_server_token(session, oauth_client_data, oauth_token_data):
    data = dict(
        token=oauth_token_data,
    )
    action = OAuthServerTokenDeleteAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.DELETE
    assert rows[0].model == ServerToken


def test_delete_oauth_server_token_w_client(
    session, oauth_client_data, oauth_token_data
):
    data = dict(
        client=oauth_client_data,
        token=oauth_token_data,
    )
    action = OAuthServerTokenDeleteAction(data)
    with pytest.raises(AssertionError):
        list(action.prepare(session))


def test_create_oauth_application(session, oauth_client_data):
    data = dict(
        client=oauth_client_data,
    )
    action = OAuthApplicationCreateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.INSERT
    assert rows[0].model == ServerClient


def test_update_oauth_application(session, oauth_client_data):
    data = dict(
        client=oauth_client_data,
    )
    action = OAuthApplicationUpdateAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == ServerClient


def test_delete_oauth_application(session, oauth_client_data):
    data = dict(
        client=oauth_client_data,
    )
    action = OAuthApplicationDeleteAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 1
    assert rows[0].type == OperationType.DELETE
    assert rows[0].model == ServerClient


###
# OAuth Client
###


@pytest.fixture(scope="function")
def oauth_remote_account():
    """OAuth client remote account."""
    return {
        "id": 8546,
        "user_id": 22858,
        "client_id": "APP-MAX7XCD8Q98X4VT6",
        "extra_data": '{"orcid": "0000-0002-5676-5956", "full_name": "Alex Ioannidis"}',
        "created": "2023-06-29T13:00:00",
        "updated": "2023-06-29T14:00:00",
    }


@pytest.fixture(scope="function")
def oauth_remote_token():
    """OAuth client remote token."""
    return {
        "id_remote_account": 8546,
        "token_type": "",
        "access_token": "R3RVeGc3K0RrM25rbXc4ZWxGM3oxYVA4LzcwVWpCNkM4aG8vRy9CNWxkZFFCMk9OR1d2d29lN3dKdWk2eEVTQQ==",
        "secret": "",
        "created": "2023-06-29T13:00:00",
        "updated": "2023-06-29T14:00:00",
    }


@pytest.fixture(scope="function")
def account_user_identity():
    """OAuth client user identity."""
    return {
        "id": "0000-0002-5676-5956",
        "method": "orcid",
        "id_user": 22858,
        "created": "2023-06-29T13:00:00",
        "updated": "2023-06-29T14:00:00",
    }


def test_connect_minimal_oauth_account(
    session,
    oauth_remote_account,
    oauth_remote_token,
    account_user_identity,
):
    data = dict(
        remote_account=oauth_remote_account,
        remote_token=oauth_remote_token,
        user_identity=account_user_identity,
    )
    action = OAuthLinkedAccountConnectAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 3
    assert rows[0].type == OperationType.INSERT
    assert rows[0].model == RemoteAccount
    assert rows[1].type == OperationType.INSERT
    assert rows[1].model == RemoteToken
    assert rows[2].type == OperationType.INSERT
    assert rows[2].model == UserIdentity


def test_connect_full_oauth_account(
    session,
    oauth_remote_account,
    oauth_remote_token,
    account_user_identity,
    oauth_client_data,
    oauth_token_data,
):
    data = dict(
        remote_account=oauth_remote_account,
        remote_token=oauth_remote_token,
        user_identity=account_user_identity,
        server_client=oauth_client_data,
        server_token=oauth_token_data,
    )
    action = OAuthLinkedAccountConnectAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 5
    assert rows[0].type == OperationType.INSERT
    assert rows[0].model == RemoteAccount
    assert rows[1].type == OperationType.INSERT
    assert rows[1].model == RemoteToken
    assert rows[2].type == OperationType.INSERT
    assert rows[2].model == UserIdentity
    assert rows[3].type == OperationType.INSERT
    assert rows[3].model == ServerClient
    assert rows[4].type == OperationType.INSERT
    assert rows[4].model == ServerToken


def test_disconnect_oauth_account(
    session,
    oauth_remote_account,
    oauth_remote_token,
    account_user_identity,
):
    data = dict(
        remote_account=oauth_remote_account,
        remote_token=oauth_remote_token,
        user_identity=account_user_identity,
    )
    action = OAuthLinkedAccountDisconnectAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 3
    assert rows[0].type == OperationType.DELETE
    assert rows[0].model == UserIdentity
    assert rows[1].type == OperationType.DELETE
    assert rows[1].model == RemoteToken
    assert rows[2].type == OperationType.DELETE
    assert rows[2].model == RemoteAccount


def test_disconnect_gh_oauth_account(session, oauth_token_data, account_user_identity):
    data = dict(
        token=oauth_token_data,
        user_identity=account_user_identity,
    )
    action = OAuthGHDisconnectToken(data)
    rows = list(action.prepare(session))

    assert len(rows) == 2
    assert rows[0].type == OperationType.DELETE
    assert rows[0].model == UserIdentity
    assert rows[1].type == OperationType.DELETE
    assert rows[1].model == ServerToken



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""JSONLines extract."""

from pathlib import Path

import orjson

from invenio_rdm_migrator.extract import Extract


class JSONLExtract(Extract):
    """Data extraction from JSONL files."""

    def __init__(self, filepath):
        """Constructor."""
        if not Path(filepath).exists():
            raise FileNotFoundError(filepath)

        self.filepath = filepath

    def run(self):
        """Yield one element at a time."""
        with open(self.filepath, "rb") as reader:
            for line in reader:
                yield orjson.loads(line)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration OAuth load module."""

from ...load.postgresql.bulk import PostgreSQLCopyLoad
from ...load.postgresql.bulk.generators import (
    ExistingDataTableGenerator,
    SingleTableGenerator,
)
from ..models.oauth import RemoteAccount, RemoteToken, ServerClient, ServerToken


class OAuthClientCopyLoad(PostgreSQLCopyLoad):
    """OAuthClient loading of existing data."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(
            table_generators=[
                ExistingDataTableGenerator(tables=[RemoteAccount, RemoteToken]),
            ],
            **kwargs,
        )


class OAuthServerClientCopyLoad(PostgreSQLCopyLoad):
    """OAuth2Server clients loading."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(
            table_generators=[SingleTableGenerator(table=ServerClient)],
            **kwargs,
        )


class OAuthServerTokenCopyLoad(PostgreSQLCopyLoad):
    """OAuth2Server tokens loading."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(
            table_generators=[SingleTableGenerator(table=ServerToken)],
            **kwargs,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""State tests."""

import tempfile
from datetime import datetime
from pathlib import Path

import pytest
from sqlalchemy import MetaData, create_engine, insert, select
from sqlalchemy.exc import IntegrityError

from invenio_rdm_migrator.state import StateDB, StateValidator

###
# Global
###


@pytest.fixture(scope="function")
def disk_db(tmp_dir):
    """Creates a DB on disk and initializes all the migrator related tables on it."""
    db_dir = Path(tmp_dir.name) / "state.db"
    disk_eng = create_engine(f"sqlite:///{db_dir}")
    disk_meta = MetaData()
    StateDB._initialize_db(disk_meta)
    disk_meta.create_all(disk_eng)
    with disk_eng.connect() as conn:
        conn.execute(
            insert(disk_meta.tables["parents"]).values(
                {
                    "recid": "123456",
                    "id": "1234abcd-1234-5678-abcd-123abc456def",
                    "latest_id": "12345678-abcd-1a2b-3c4d-123abc456def",
                    "latest_index": 1,
                },
            )
        )
        conn.execute(
            insert(disk_meta.tables["records"]).values(
                {
                    "recid": "123456",
                    "index": 1,
                    "id": "1234abcd-1234-5678-abcd-123abc456def",  # uuid
                    "parent_id": "1234abcd-1234-5678-abcd-123abc456def",  # parent uuid
                    "fork_version_id": 3,
                    "pids": {"doi": "10.1234/123456"},
                },
            )
        )
        conn.execute(
            insert(disk_meta.tables["communities"]).values(
                {
                    "slug": "test-comm",
                    "id": "1234abcd-1234-5678-abcd-123abc456def",
                    "bucket_id": "1234abcd-1234-5678-abcd-123abc456def",
                    "oai_set_id": 1,
                }
            )
        )
        conn.execute(
            insert(disk_meta.tables["global"]).values(
                {"key": "max_value_pk", "value": 1}
            )
        )

        conn.commit()

    yield disk_eng, disk_meta

    db_dir.unlink()


def test_state_with_no_initial_db(tmp_dir):
    """Test state creation without an existing db.

    Test the implicit creation of tables.
    """
    state = StateDB(db_dir=tmp_dir.name)
    # tables exist and are empty
    assert len(list(state.all("parents"))) == 0
    assert len(list(state.all("records"))) == 0
    assert len(list(state.all("communities"))) == 0
    assert len(list(state.all("pids"))) == 0
    assert len(list(state.all("global"))) == 0

    # check that no extra tables were created
    assert set(state.tables) == set(
        {
            "parents",
            "records",
            "communities",
            "global",
            "pids",
            "buckets",
            "file_records",
        }
    )


def test_state_load_from_disk(tmp_dir, disk_db):
    """Test db loading from disk into memory."""
    state = StateDB(db_dir=tmp_dir.name)

    # check the memory db status
    assert len(list(state.all("parents"))) == 1
    assert len(list(state.all("records"))) == 1
    assert len(list(state.all("communities"))) == 1
    assert len(list(state.all("global"))) == 1

    # add data to disk db and check they are not present in memory
    state.add(
        "parents",
        {
            "recid": "123455",
            "id": "1234abbb-1234-5678-abcd-123abc456def",
            "latest_id": "12345678-abcd-1a2b-3c4d-123abc456def",
            "latest_index": 1,
        },
    )
    assert len(list(state.all("parents"))) == 2

    # check the db on disk has not been updated (prevent pollution)
    disk_eng, disk_meta = disk_db
    parents_table = disk_meta.tables["parents"]
    with disk_eng.connect() as conn:
        assert (
            len(
                list(
                    conn.execute(
                        select(parents_table).where(
                            parents_table.columns.recid == "123455"
                        )
                    ).all()
                )
            )
            == 0
        )


def test_state_save_to_disk(tmp_dir):
    """Test db persistence to disk."""
    state = StateDB(db_dir=tmp_dir.name)
    # check the memory db status
    assert len(list(state.all("parents"))) == 0
    assert len(list(state.all("records"))) == 0
    assert len(list(state.all("communities"))) == 0
    assert len(list(state.all("global"))) == 0

    # add data to disk db and check they are not present in memory
    state.add(
        "parents",
        {
            "recid": "123455",
            "id": "1234abbb-1234-5678-abcd-123abc456def",
            "latest_id": "12345678-abcd-1a2b-3c4d-123abc456def",
            "latest_index": 1,
        },
    )
    assert len(list(state.all("parents"))) == 1
    state.save()

    # check if data is saved to disk
    db_dir = Path(tmp_dir.name) / "state.db"
    disk_eng = create_engine(f"sqlite:///{db_dir}")
    disk_meta = MetaData()
    _ = StateDB._initialize_db(disk_meta)
    disk_meta.create_all(disk_eng)

    parents_table = disk_meta.tables["parents"]
    with disk_eng.connect() as conn:
        assert (
            len(
                list(
                    conn.execute(
                        select(parents_table).where(
                            parents_table.columns.recid == "123455"
                        )
                    ).all()
                )
            )
            == 1
        )


def test_state_save_to_disk_alt_filename(tmp_dir):
    """Test db persistence to disk with a different filename."""
    state = StateDB(db_dir=tmp_dir.name)
    state.save(filename="altname.db")
    assert (Path(tmp_dir.name) / "altname.db").exists()


def test_state_save_to_disk_alt_filepath(tmp_dir):
    """Test db persistence to disk on a different path."""
    state = StateDB(db_dir=tmp_dir.name)

    alt_dir = tempfile.TemporaryDirectory()
    alt_path = Path(alt_dir.name) / "altpath.db"

    state.save(filepath=alt_path)
    assert alt_path.exists()

    alt_dir.cleanup()


###
# Parent State
# record uuid = d3c0dd1d...
# parent uuid = f4d3071d...
###


def test_parent_state_record_entry(parents_state):
    parents_state.PARENTS.add(
        "123",
        {
            "id": "f4d3071d-1234-abcd-1ab2-1234abcd56ef",
            "latest_index": 1,
            "latest_id": "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef",
        },
    )
    assert parents_state.PARENTS.get("123")


def test_parent_state_draft_entry(parents_state):
    parents_state.PARENTS.add(
        "123",
        {
            "id": "f4d3071d-1234-abcd-1ab2-1234abcd56ef",
            "next_draft_id": "dd4f71d0-1234-abcd-1ab2-1234abcd56ef",
        },
    )
    assert parents_state.PARENTS.get("123")


def test_parent_state_update_record(parents_state):
    parents_state.PARENTS.add(
        "123",
        {
            "id": "f4d3071d-1234-abcd-1ab2-1234abcd56ef",
            "latest_index": 1,
            "latest_id": "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef",
        },
    )
    assert not parents_state.PARENTS.get("123").get("next_draft_id")

    parents_state.PARENTS.update(
        "123",
        {
            "latest_index": 100,
            "latest_id": "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef",
        },
    )
    assert parents_state.PARENTS.get("123").get("latest_index") == 100
    assert (
        parents_state.PARENTS.get("123").get("latest_id")
        == "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef"
    )


def test_parent_state_update_draft(parents_state):
    parents_state.PARENTS.add(
        "123",
        {
            "id": "f4d3071d-1234-abcd-1ab2-1234abcd56ef",
            "latest_index": 1,
            "latest_id": "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef",
        },
    )
    assert not parents_state.PARENTS.get("123").get("next_draft_id")

    parents_state.PARENTS.update(
        "123",
        {
            "next_draft_id": "dd4f71d0-1234-abcd-1ab2-1234abcd56ef",
        },
    )
    assert (
        parents_state.PARENTS.get("123").get("next_draft_id")
        == "dd4f71d0-1234-abcd-1ab2-1234abcd56ef"
    )


def test_parent_state_invalid_entries(parents_state):
    invalid_by_db_constraints = [
        {  # no id
            "latest_index": 1,
            "latest_id": "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef",
        },
    ]

    for entry in invalid_by_db_constraints:
        pytest.raises(IntegrityError, parents_state.PARENTS.add, "123", entry)

    invalid_by_validator = [
        {  # missing lastest index
            "id": "f4d3071d-1234-abcd-1ab2-1234abcd56ef",
            "latest_id": "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef",
        },
        {  # no latest id nor next draft id
            "id": "f4d3071d-1234-abcd-1ab2-1234abcd56ef",
            "latest_index": 1,
        },
    ]

    for entry in invalid_by_validator:
        pytest.raises(AssertionError, parents_state.PARENTS.add, "123", entry)


###
# Record State
###


def test_record_state_record_entry(state):
    state.RECORDS.add(
        "123",
        {
            "index": 1,
            "id": "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef",
            "parent_id": "f4d3071d-1234-abcd-1ab2-1234abcd56ef",
            "fork_version_id": 1,
        },
    )
    assert state.RECORDS.get("123")


def test_record_state_invalid_entries(state):
    invalid = [
        {  # missing index
            "id": "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef",
            "parent_id": "f4d3071d-1234-abcd-1ab2-1234abcd56ef",
            "fork_version_id": 1,
        },
        {  # missing id
            "index": 1,
            "parent_id": "f4d3071d-1234-abcd-1ab2-1234abcd56ef",
            "fork_version_id": 1,
        },
        {  # missing parent_id
            "index": 1,
            "id": "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef",
            "fork_version_id": 1,
        },
        {  # missing fork_version_id
            "index": 1,
            "id": "d3c0dd1d-1234-abcd-1ab2-1234abcd56ef",
            "parent_id": "f4d3071d-1234-abcd-1ab2-1234abcd56ef",
        },
    ]

    for entry in invalid:
        pytest.raises(IntegrityError, state.RECORDS.add, "123", entry)


###
# PIDs State
###


def test_pids_state_valid(state):
    created = datetime.fromtimestamp(1688045928)
    state.PIDS.add(
        "123",
        {
            "id": 1_000_000,
            "pid_type": "recid",
            "status": "K",
            "created": created,
        },
    )
    state.PIDS.add(
        "124",
        {
            "id": 1000001,
            "pid_type": "recid",
            "status": "K",
            "obj_type": "rec",
            "created": created,
        },
    )

    assert state.PIDS.get("123")
    assert state.PIDS.get("124")


def test_pids_state_duplicated_id(state):
    created = datetime.fromtimestamp(1688045928)
    state.PIDS.add(
        "123",
        {
            "id": 1_000_000,  # has a unique constraint
            "pid_type": "recid",
            "status": "K",
            "created": created,
        },
    )

    pytest.raises(
        IntegrityError,
        state.PIDS.add,
        "124",  # diff pk to make sure it fails on id
        {
            "id": 1_000_000,  # has a unique constraint
            "pid_type": "recid",
            "status": "K",
            "created": created,
        },
    )


def test_pids_state_invalid_entries(state):
    created = datetime.fromtimestamp(1688045928)
    invalid = [
        {  # missing id
            "pid_type": "recid",
            "status": "K",
            "obj_type": "rec",
            "created": created,
        },
        {  # missing pid_type
            "id": 1_000_000,
            "status": "K",
            "obj_type": "rec",
            "created": created,
        },
        {  # missing status
            "id": 1_000_000,
            "pid_type": "recid",
            "obj_type": "rec",
            "created": created,
        },
        {  # missing created
            "id": 1_000_000,
            "status": "K",
            "pid_type": "recid",
            "obj_type": "rec",
        },
    ]

    for idx, entry in enumerate(invalid):
        pytest.raises(IntegrityError, state.PIDS.add, idx, entry)


###
# StateValidator
###


def test_default_validator():
    # test that the interface holds
    assert StateValidator.validate({}) == False


###
# Bucket State
###


def test_bucket_state_valid(state):
    state.BUCKETS.add(
        "0e12b4b6-9cc7-46df-9a04-c11c478de211",
        {"draft_id": "d94f793c-47d2-48e2-9867-ca597b4ebb41"},
    )

    assert state.BUCKETS.get("0e12b4b6-9cc7-46df-9a04-c11c478de211")


def test_buckets_state_invalid_entries(state):
    pytest.raises(
        IntegrityError,
        state.BUCKETS.add,
        "0e12b4b6-9cc7-46df-9a04-c11c478de211",
        {},  # empty
    )


def test_delete_bucket_state(state):
    state.BUCKETS.add(
        "0e12b4b6-9cc7-46df-9a04-c11c478de211",
        {"draft_id": "d94f793c-47d2-48e2-9867-ca597b4ebb41"},
    )

    state.BUCKETS.delete("0e12b4b6-9cc7-46df-9a04-c11c478de211")
    assert not state.BUCKETS.get("0e12b4b6-9cc7-46df-9a04-c11c478de211")


###
# File Records
###


def test_file_record_state(state):
    state.FILE_RECORDS.add(
        "e94b243e-9c0c-44df-bd1f-6decc374cf78",
        {
            "json": {},
            "created": "2023-06-29T13:00:00",
            "updated": "2023-06-29T14:00:00",
            "version_id": 1,
            "key": "IMG_3535.jpg",
            "record_id": "93c09d1d-47d2-48e2-9867-ca597b4ebb41",
            "object_version_id": "f8200dc7-55b6-4785-abd0-f3d13b143c98",
        },
    )

    assert state.FILE_RECORDS.get("e94b243e-9c0c-44df-bd1f-6decc374cf78")


def test_file_record_state_search(state):
    state.FILE_RECORDS.add(
        "e94b243e-9c0c-44df-bd1f-6decc374cf78",
        {
            "json": {},
            "created": "2023-06-29T13:00:00",
            "updated": "2023-06-29T14:00:00",
            "version_id": 1,
            "key": "IMG_3535.jpg",
            "record_id": "93c09d1d-47d2-48e2-9867-ca597b4ebb41",
            "object_version_id": "f8200dc7-55b6-4785-abd0-f3d13b143c98",
        },
    )
    state.FILE_RECORDS.add(
        "e94b243e-9c0c-44df-bd1f-6decc374cf79",
        {
            "json": {},
            "created": "2023-06-29T13:00:00",
            "updated": "2023-06-29T14:00:00",
            "version_id": 1,
            "key": "IMG_3535.jpg",
            "record_id": "93c09d1d-47d2-48e2-9867-ca597b4ebb41",
            "object_version_id": "f8200dc7-55b6-4785-abd0-f3d13b143c98",
        },
    )

    files = state.FILE_RECORDS.search(
        "record_id", "93c09d1d-47d2-48e2-9867-ca597b4ebb41"
    )
    assert len(list(files)) == 2



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Ignored actions module."""

from ....actions import LoadAction


class AnyData:
    """Utility class for accepting any combination of data arguments."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        pass


class IgnoredAction(LoadAction):
    """Ingored action.

    This is a utility class used in case there are actions that need to be parsed, but
    are ignored/skipped and do not translate to a set of generate rows.
    """

    name = "ignored"
    data_cls = AnyData

    def _generate_rows(self, **kwargs):
        """Yields nothing."""
        yield from iter(())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration transaction transform."""

from abc import ABC

from ..actions.base import TransformAction
from .base import Transform
from .errors import MultipleActionMatches, NoActionMatch


class BaseTxTransform(Transform, ABC):
    """Transform a transaction."""

    actions: list[TransformAction] = []

    def _detect_action(self, tx):
        match_classes = []
        for action_cls in self.actions:
            if action_cls.matches_action(tx):
                match_classes.append(action_cls)

        if len(match_classes) == 0:
            raise NoActionMatch(tx)
        elif len(match_classes) > 1:
            raise MultipleActionMatches(tx, match_classes)

        return match_classes[0]  # return the one and only matched class

    def _transform(self, tx):
        """Transform action.

        Each action contains the raw transaction data.
        :returns: an instance of LoadAction.
        """
        action_cls = self._detect_action(tx)
        action = action_cls(tx)
        return action.transform()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Dataclasses models to generate table rows for OAI-PMH."""

from dataclasses import InitVar
from datetime import datetime
from uuid import UUID

from sqlalchemy.orm import Mapped, mapped_column

from ...load.postgresql.models import Model


class WebhookEvent(Model):
    """Webhook event dataclass model."""

    __tablename__: InitVar[str] = "webhooks_events"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    created: Mapped[datetime]
    updated: Mapped[datetime]
    receiver_id: Mapped[str]
    user_id: Mapped[int] = mapped_column(nullable=True)
    payload: Mapped[dict] = mapped_column(nullable=True)
    payload_headers: Mapped[dict] = mapped_column(nullable=True)
    response: Mapped[dict] = mapped_column(nullable=True)
    response_headers: Mapped[dict] = mapped_column(nullable=True)
    response_code: Mapped[int] = mapped_column(nullable=True)


class Repository(Model):
    """Repository dataclass model."""

    __tablename__: InitVar[str] = "github_repositories"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    created: Mapped[datetime]
    updated: Mapped[datetime]
    github_id: Mapped[int] = mapped_column(nullable=True)
    name: Mapped[str]
    user_id: Mapped[int] = mapped_column(nullable=True)
    hook: Mapped[int] = mapped_column(nullable=True)


class Release(Model):
    """Release dataclass model."""

    __tablename__: InitVar[str] = "github_releases"

    created: Mapped[datetime]
    updated: Mapped[datetime]
    id: Mapped[UUID] = mapped_column(primary_key=True)
    release_id: Mapped[int] = mapped_column(nullable=True)
    tag: Mapped[str] = mapped_column(nullable=True)
    errors: Mapped[dict] = mapped_column(nullable=True)
    repository_id: Mapped[UUID] = mapped_column(nullable=True)
    event_id: Mapped[UUID] = mapped_column(nullable=True)
    record_id: Mapped[UUID] = mapped_column(nullable=True)
    status: Mapped[str]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""File actions tests."""

from datetime import datetime
from uuid import uuid4

import pytest
import sqlalchemy as sa
from testutils import assert_model_count

from invenio_rdm_migrator.load.postgresql.transactions.operations import OperationType
from invenio_rdm_migrator.streams.actions.load import FileUploadAction
from invenio_rdm_migrator.streams.models.files import (
    FilesBucket,
    FilesInstance,
    FilesObjectVersion,
)
from invenio_rdm_migrator.streams.models.records import (
    RDMDraftFile,
    RDMDraftMediaFile,
    RDMDraftMetadata,
)


@pytest.fixture(scope="function")
def draft(
    session,
    database,
    bucket_data_w_size,
):
    """Draft with bucket."""
    session.add(FilesBucket(**bucket_data_w_size))
    session.add(
        RDMDraftMetadata(
            id="5cbdc04b-b2ab-4d25-b21c-3ce6a5269710",
            json={},
            created=datetime.utcnow(),
            updated=datetime.utcnow(),
            version_id=1,
            index=1,
            bucket_id=bucket_data_w_size["id"],
            parent_id=uuid4(),
            expires_at=None,
            fork_version_id=None,
            media_bucket_id=None,
        )
    )
    session.commit()


@pytest.fixture(scope="function")
def draft_with_file(session, draft, fi_data, ov_data, fr_data):
    """Draft with one file."""
    session.add(FilesInstance(**fi_data))
    session.add(FilesObjectVersion(**ov_data))
    fr_data["id"] = uuid4()
    fr_data["record_id"] = "5cbdc04b-b2ab-4d25-b21c-3ce6a5269710"
    session.add(RDMDraftFile(**fr_data))
    session.commit()


def test_upload_file(
    session,
    database,
    pg_tx,
    draft,
    bucket_data_w_size,
    fi_data,
    ov_data,
    fr_data,
):
    """Test draft file upload replacement."""
    data = dict(
        bucket=bucket_data_w_size,
        object_version=ov_data,
        file_instance=fi_data,
        file_record=fr_data,
    )
    action = FileUploadAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 4
    bucket, fi, ov, draft_file = rows
    assert bucket.type == OperationType.UPDATE
    assert bucket.model == FilesBucket
    assert fi.type == OperationType.INSERT
    assert fi.model == FilesInstance
    assert ov.type == OperationType.INSERT
    assert ov.model == FilesObjectVersion
    assert draft_file.type == OperationType.INSERT
    assert draft_file.model == RDMDraftFile
    # still type str since it has not been inserted in DB
    assert str(draft_file.data["record_id"]) == "5cbdc04b-b2ab-4d25-b21c-3ce6a5269710"
    assert str(draft_file.data["object_version_id"]) == ov_data["version_id"]

    pg_tx.run([action])
    assert_model_count(session, FilesBucket, 1)
    assert_model_count(session, FilesObjectVersion, 1)
    assert_model_count(session, FilesInstance, 1)
    assert_model_count(session, RDMDraftFile, 1)


def test_upload_file_replacement(
    session,
    database,
    pg_tx,
    draft_with_file,
    bucket_data_w_size,
    fi_data,
    ov_data,
    fr_data,
):
    """Test draft file upload replacement."""
    fi_data["id"] = str(uuid4())
    ov_data["file_id"] = fi_data["id"]
    ov_data["version_id"] = str(uuid4())
    data = dict(
        bucket=bucket_data_w_size,
        object_version=ov_data,
        file_instance=fi_data,
        file_record=fr_data,
    )
    action = FileUploadAction(data)
    rows = list(action.prepare(session))

    assert len(rows) == 5
    bucket, old_ov, fi, new_ov, draft_file = rows
    assert bucket.type == OperationType.UPDATE
    assert bucket.model == FilesBucket
    assert old_ov.type == OperationType.UPDATE
    assert old_ov.model == FilesObjectVersion
    assert str(old_ov.data["version_id"]) == "f8200dc7-55b6-4785-abd0-f3d13b143c98"
    assert fi.type == OperationType.INSERT
    assert fi.model == FilesInstance
    assert new_ov.type == OperationType.INSERT
    assert new_ov.model == FilesObjectVersion
    assert draft_file.type == OperationType.UPDATE
    assert draft_file.model == RDMDraftFile
    assert str(draft_file.data["record_id"]) == "5cbdc04b-b2ab-4d25-b21c-3ce6a5269710"
    assert str(draft_file.data["object_version_id"]) == ov_data["version_id"]

    pg_tx.run([action])
    assert_model_count(session, FilesBucket, 1)
    assert_model_count(session, FilesObjectVersion, 2)
    assert_model_count(session, FilesInstance, 2)
    assert_model_count(session, RDMDraftFile, 1)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Funders models."""

from dataclasses import InitVar
from datetime import datetime
from uuid import UUID

from sqlalchemy.orm import Mapped, mapped_column

from ...load.postgresql.models import Model


class Funders(Model):
    """Funders dataclass model."""

    __tablename__: InitVar[str] = "funder_metadata"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    pid: Mapped[str]
    json: Mapped[dict]
    created: Mapped[datetime]
    updated: Mapped[datetime]
    version_id: Mapped[int]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration deleted record load."""

from datetime import datetime

from ....load.ids import generate_recid, generate_uuid, pid_pk
from ....load.postgresql.bulk.generators import TableGenerator
from ....state import STATE
from ...models.pids import PersistentIdentifier
from ...models.records import RDMParentMetadata, RDMRecordMetadata


def generate_record_uuid(data):
    """Generate record uuid if not present."""
    _id = data["record"].get("id")
    return _id if _id else generate_uuid(None)


class RDMDeletedRecordTableGenerator(TableGenerator):
    """RDM Record and related tables load."""

    def __init__(self):
        """Constructor."""
        super().__init__(
            tables=[
                PersistentIdentifier,
                RDMParentMetadata,
                RDMRecordMetadata,
            ],
            pks=[
                ("record.id", generate_record_uuid),
                ("parent.id", generate_uuid),
                ("record.json.pid", generate_recid),
                ("parent.json.pid", generate_recid),
                ("record.parent_id", lambda d: d["parent"]["id"]),
            ],
        )

    def _generate_rows(self, data, **kwargs):
        """Generates rows for a record."""
        now = datetime.utcnow().isoformat()
        parent = data["parent"]
        record = data["record"]

        # Handle parent
        parent_pid_value = parent["json"]["id"]
        state_parent = STATE.PARENTS.get(parent_pid_value)
        if not state_parent:
            if parent_pid_value:
                STATE.PARENTS.add(
                    parent["json"]["id"],  # recid
                    {
                        "id": parent["id"],
                        "latest_index": record["index"],
                        "latest_id": record["id"],
                        "communities": parent.get("communities", {}).get("ids", []),
                    },
                )
                parent_pid = parent["json"]["pid"]
                yield PersistentIdentifier(
                    id=parent_pid["pk"],
                    pid_type=parent_pid["pid_type"],
                    pid_value=parent["json"]["id"],
                    status=parent_pid["status"],
                    object_type=parent_pid["obj_type"],
                    object_uuid=parent["id"],
                    created=now,
                    updated=now,
                )
            parent_doi = parent["json"].get("pids", {}).get("doi")
            if parent_doi and parent_doi["identifier"]:
                yield PersistentIdentifier(
                    id=pid_pk(),
                    pid_type="doi",
                    pid_value=parent_doi["identifier"],
                    status="R",
                    object_type="rec",
                    object_uuid=parent["id"],
                    created=now,
                    updated=now,
                )

            # parent record
            yield RDMParentMetadata(
                id=parent["id"],
                json=parent["json"],
                created=parent["created"],
                updated=parent["updated"],
                version_id=parent["version_id"],
            )
        parent_id = state_parent["id"] if state_parent else record["parent_id"]

        # record
        record_state = STATE.RECORDS.get(record["json"]["id"])
        if not record_state:
            STATE.RECORDS.add(
                record["json"]["id"],  # recid
                {
                    "index": record["index"],
                    "id": record["id"],  # uuid
                    "parent_id": parent_id,  # parent uuid
                    "fork_version_id": record["version_id"],
                    "pids": record["json"].get("pids", {}),
                },
            )
        else:
            # Something is very wrong, we bail out
            return

        yield RDMRecordMetadata(
            id=record["id"],
            json=record["json"],
            created=record["created"],
            updated=record["updated"],
            version_id=record["version_id"],
            index=record["index"],
            bucket_id=record.get("bucket_id"),
            media_bucket_id=record.get("media_bucket_id"),
            parent_id=parent_id,
            deletion_status="D",
        )
        # recid
        record_pid = record["json"]["pid"]
        yield PersistentIdentifier(
            id=record_pid["pk"],
            pid_type=record_pid["pid_type"],
            pid_value=record["json"]["id"],
            status=record_pid["status"],
            object_type=record_pid["obj_type"],
            object_uuid=record["id"],
            created=now,
            updated=now,
        )
        # DOI
        if "doi" in record["json"].get("pids", {}):
            yield PersistentIdentifier(
                id=pid_pk(),
                pid_type="doi",
                pid_value=record["json"]["pids"]["doi"]["identifier"],
                status="R",
                object_type="rec",
                object_uuid=record["id"],
                created=now,
                updated=now,
            )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Communities load tests."""

import os
from unittest.mock import patch

import pytest

from invenio_rdm_migrator.streams.communities import CommunityCopyLoad


@pytest.fixture(scope="function")
def community_copy_load(state, tmp_dir):
    """Community load instance."""
    # the db queries will be mocked
    load = CommunityCopyLoad(db_uri="None", tmp_dir=tmp_dir.name)
    yield load
    load._cleanup()


@patch(
    "invenio_rdm_migrator.load.ids.uuid4",
    lambda: "7357c033-abcd-1a2b-3c4d-123abc456def",
)
def test_community_load_prepare(community_copy_load, transformed_community_entry):
    """Test the table preparation (file creation)."""
    tables = list(community_copy_load._prepare([transformed_community_entry]))

    # assert tables
    # it has four tables: communities, parent communities, communities_files and members
    expected__tablename__s = [
        "community_metadata",
        "oaiserver_set",
        "communities_members",
        "rdm_parents_community",
        "communities_files",
        "files_object",
        "files_bucket",
    ]
    table_names = [table.__tablename__ for _, table in tables]
    assert len(tables) == len(expected__tablename__s)
    assert any([e_table in table_names for e_table in expected__tablename__s])

    # assert files were created and have the content
    # five files are created: one for the communities, members, files, files bucket, files object
    files = list(os.scandir(community_copy_load.tmp_dir))
    assert len(files) == 6

    # assert communities metadata content
    with open(f"{community_copy_load.tmp_dir}/communities_metadata.csv", "r") as file:
        # uuid, json, created, updated, version_id, number, expired_at
        expected = (
            "7357c033-abcd-1a2b-3c4d-123abc456def,"
            '2023-01-01 12:00:00.00000,2023-01-31 12:00:00.00000,"{""title"":""Migrator community"",""description"":""Migrator testing community"",""page"":"""",""curation_policy"":""""}",1,migrator,7357c033-abcd-1a2b-3c4d-123abc456def,P\n'
        )
        content = file.read()
        assert content == expected

    # assert communities members content
    with open(f"{community_copy_load.tmp_dir}/communities_members.csv", "r") as file:
        # uuid, json, created, updated, version_id, number, expired_at
        expected = "7357c033-abcd-1a2b-3c4d-123abc456def,2023-01-01 12:00:00.00000,2023-01-31 12:00:00.00000,{},1,owner,True,True,7357c033-abcd-1a2b-3c4d-123abc456def,1,,\n"
        content = file.read()
        assert content == expected



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""JSON field load module."""

from datetime import datetime


class DatetimeMixin:
    """Transform integers to datetime.

    Assumes no timezone (UTC).
    """

    @staticmethod
    def _microseconds_to_isodate(data, fields):
        """Transform int to datetime."""
        for field in fields:
            value = data.get(field)
            if isinstance(value, int):  # would also ignore None values
                data[field] = datetime.utcfromtimestamp(value / 1_000_000).isoformat()

    @staticmethod
    def _milliseconds_to_isodate(data, fields):
        """Transform int to datetime."""
        for field in fields:
            value = data.get(field)
            if isinstance(value, int):  # would also ignore None values
                data[field] = datetime.utcfromtimestamp(value / 1_000).isoformat()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration files table load module."""

from ...load.postgresql.bulk.generators import TableGenerator
from ..models.files import FilesBucket, FilesInstance, FilesObjectVersion


class FilesTableGenerator(TableGenerator):
    """Files and related tables load via CSV."""

    def __init__(self):
        """Constructor."""
        super().__init__(
            tables=[
                FilesInstance,
                FilesBucket,
                FilesObjectVersion,
            ],
        )

    def _generate_rows(self, data, **kwargs):
        bucket = data["bucket"]
        object_version = data["object_version"]
        file = data["file"]
        yield FilesInstance(**file)
        yield FilesBucket(**bucket)
        yield FilesObjectVersion(**object_version)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Names stream."""

from .load import ExistingNamesLoad

__all__ = ("ExistingNamesLoad",)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration load actions."""

from .communities import (
    CommunityCreateAction,
    CommunityDeleteAction,
    CommunityUpdateAction,
)
from .drafts import (
    DraftCreateAction,
    DraftEditAction,
    DraftPublishEditAction,
    DraftPublishNewAction,
)
from .files import (
    FileDeleteAction,
    FileUploadAction,
    MediaFileDeleteAction,
    MediaFileUploadAction,
)
from .github import (
    HookEventCreateAction,
    HookEventUpdateAction,
    ReleaseProcessAction,
    ReleaseReceiveAction,
    ReleaseUpdateAction,
    RepoCreateAction,
    RepoUpdateAction,
)
from .ignored import IgnoredAction
from .oauth import (
    OAuthApplicationCreateAction,
    OAuthApplicationDeleteAction,
    OAuthApplicationUpdateAction,
    OAuthGHDisconnectToken,
    OAuthLinkedAccountConnectAction,
    OAuthLinkedAccountDisconnectAction,
    OAuthServerTokenCreateAction,
    OAuthServerTokenDeleteAction,
    OAuthServerTokenUpdateAction,
)
from .users import UserDeactivationAction, UserEditAction, UserRegistrationAction

__all__ = (
    "CommunityCreateAction",
    "CommunityDeleteAction",
    "CommunityUpdateAction",
    "DraftCreateAction",
    "DraftEditAction",
    "FileUploadAction",
    "FileDeleteAction",
    "MediaFileUploadAction",
    "MediaFileDeleteAction",
    "DraftPublishNewAction",
    "DraftPublishEditAction",
    "IgnoredAction",
    "HookEventCreateAction",
    "HookEventUpdateAction",
    "RepoCreateAction",
    "RepoUpdateAction",
    "OAuthApplicationCreateAction",
    "OAuthApplicationDeleteAction",
    "OAuthApplicationUpdateAction",
    "OAuthGHDisconnectToken",
    "OAuthLinkedAccountConnectAction",
    "OAuthLinkedAccountDisconnectAction",
    "OAuthServerTokenCreateAction",
    "OAuthServerTokenDeleteAction",
    "OAuthServerTokenUpdateAction",
    "ReleaseReceiveAction",
    "ReleaseUpdateAction",
    "ReleaseProcessAction",
    "UserDeactivationAction",
    "UserEditAction",
    "UserRegistrationAction",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration streams."""

from datetime import datetime

from ..extract import NullExtract
from ..logging import Logger
from ..transform import IdentityTransform


class StreamDefinition:
    """ETL stream definition.

    It defines the classes that will form the ETL stream.
    """

    def __init__(self, name, extract_cls, transform_cls, load_cls):
        """Constructor."""
        self.name = name
        self.extract_cls = extract_cls
        self.transform_cls = transform_cls
        self.load_cls = load_cls


class Stream:
    """ETL stream."""

    def __init__(self, name, extract, transform, load):
        """Constructor."""
        self.name = name
        self.extract = extract or NullExtract()
        self.transform = transform or IdentityTransform()
        self.load = load

    def run(self, cleanup=False):
        """Run ETL stream."""
        logger = Logger.get_logger()

        start_time = datetime.now()
        logger.info(f"Stream {self.name} started {start_time.isoformat()}")

        extract_gen = self.extract.run()
        transform_gen = self.transform.run(extract_gen)
        self.load.run(transform_gen, cleanup=cleanup)

        end_time = datetime.now()
        logger.info(f"Stream ended {end_time.isoformat()}")

        logger.info(f"Execution time: {end_time - start_time}")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Requests table generator tests."""

from copy import deepcopy

import dictdiffer

from invenio_rdm_migrator.streams.models.requests import RequestMetadata
from invenio_rdm_migrator.streams.requests.load import RequestTableGenerator


def test_generate_rows(communities_state, transformed_incl_req_entry_pks):
    """Test the row generation of the request table generator."""
    tg = RequestTableGenerator()  # no need for state in this test
    rows = list(tg._generate_rows(transformed_incl_req_entry_pks))
    expected_rows = [
        RequestMetadata(
            id="12345678-abcd-1a2b-3c4d-123abc456def",
            json={
                "type": "community-inclusion",
                "title": "title",
                "topic": {"record": "123456"},
                "status": "submitted",
                "receiver": {"community": "comm"},
                "created_by": {"user": "3"},
                "$schema": "local://requests/request-v1.0.0.json",
            },
            created="01/01/2023",
            updated="01/01/2023",
            version_id="1",
            number="1",
            expires_at="01/01/2023",
        ),
    ]

    assert rows == expected_rows


def test_resolve_references(communities_state, transformed_incl_req_entry_pks):
    """Test the parent and community reference resolution."""
    expected = deepcopy(transformed_incl_req_entry_pks)
    expected["json"]["receiver"]["community"] = "12345678-abcd-1a2b-3c4d-123abc456def"

    tg = RequestTableGenerator()
    tg._resolve_references(transformed_incl_req_entry_pks)  # changes in place

    assert not list(dictdiffer.diff(transformed_incl_req_entry_pks, expected))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""OAuth client applications actions module."""

from dataclasses import dataclass
from typing import Optional

from .....actions import LoadAction, LoadData
from .....load.postgresql.transactions.operations import Operation, OperationType
from ....models.oauth import RemoteAccount, RemoteToken, ServerClient, ServerToken
from ....models.users import UserIdentity


@dataclass
class OAuthLinkedAccountData(LoadData):
    """OAuth application action data."""

    remote_account: dict
    remote_token: dict
    user_identity: Optional[dict] = None
    server_token: Optional[dict] = None
    server_client: Optional[dict] = None


class OAuthLinkedAccountConnectAction(LoadAction):
    """Connect an external account."""

    name = "oauth-application-connect"
    data_cls = OAuthLinkedAccountData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new oauth token."""
        assert self.data.user_identity

        yield Operation(OperationType.INSERT, RemoteAccount, self.data.remote_account)
        yield Operation(OperationType.INSERT, RemoteToken, self.data.remote_token)
        yield Operation(OperationType.INSERT, UserIdentity, self.data.user_identity)

        if self.data.server_client:
            yield Operation(OperationType.INSERT, ServerClient, self.data.server_client)
        if self.data.server_token:
            yield Operation(OperationType.INSERT, ServerToken, self.data.server_token)


class OAuthLinkedAccountDisconnectAction(LoadAction):
    """Disonnect an external account."""

    name = "oauth-application-disconnect"
    data_cls = OAuthLinkedAccountData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new oauth token."""
        if self.data.user_identity:  # GH discconnet does not have it
            yield Operation(OperationType.DELETE, UserIdentity, self.data.user_identity)

        yield Operation(OperationType.DELETE, RemoteToken, self.data.remote_token)
        yield Operation(OperationType.DELETE, RemoteAccount, self.data.remote_account)


@dataclass
class OAuthGHTokenDisconnectData(LoadData):
    """OAuth GH disconnect token data."""

    token: dict
    user_identity: dict


class OAuthGHDisconnectToken(LoadAction):
    """Disconnect GH linked account server token and identity."""

    name = "oauth-gh-application-disconnect"
    data_cls = OAuthGHTokenDisconnectData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new oauth token."""
        yield Operation(OperationType.DELETE, UserIdentity, self.data.user_identity)
        yield Operation(OperationType.DELETE, ServerToken, self.data.token)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Requests load tests."""

import os
from unittest.mock import patch

import pytest

from invenio_rdm_migrator.streams.files import FilesCopyLoad


@pytest.fixture(scope="function")
def files_copy_load(tmp_dir):
    """Request load instance."""
    # the db queries will be mocked
    load = FilesCopyLoad(db_uri="None", tmp_dir=tmp_dir.name)
    yield load
    load._cleanup()


@patch(
    "invenio_rdm_migrator.load.ids.uuid4",
    lambda: "12345678-abcd-1a2b-3c4d-123abc456def",
)
def test_files_load_prepare(files_copy_load, transformed_files_entry):
    """Test the table preparation (file creation)."""
    tables = [table for _, table in files_copy_load._prepare([transformed_files_entry])]

    # assert tables
    assert len(tables) == 3
    assert tables[0].__tablename__ == "files_files"
    assert tables[1].__tablename__ == "files_bucket"
    assert tables[2].__tablename__ == "files_object"

    # assert files were created and have the content
    files = list(os.scandir(files_copy_load.tmp_dir))
    assert len(files) == 3

    # assert files bucket content
    with open(f"{files_copy_load.tmp_dir}/files_bucket.csv", "r") as file:
        # id, created, updated, default_location, default_storage_class, size, quota_size, max_file_size,locked, deleted
        expected = (
            "1",
            "2023-04-19",
            "2023-04-19",
            "1",
            "L",
            "1234",
            "False",
            "False",
            "",
            "",
        )
        content = file.read().rstrip()
        assert content == ",".join(expected)
    # assert files object version content
    with open(f"{files_copy_load.tmp_dir}/files_files.csv", "r") as file:
        # id, created, updated, uri, storage_class, size, checksum, readable, writable, last_check_at, last_check
        expected = (
            "3",
            "2023-04-19",
            "2023-04-19",
            "path/to/file",
            "L",
            "1234",
            "md5:abcd1234",
            "True",
            "True",
            "",
            "False",
        )
        content = file.read().rstrip()
        assert content == ",".join(expected)
    # assert files instance content
    with open(f"{files_copy_load.tmp_dir}/files_object.csv", "r") as file:
        #  version_id, created, updated, key, bucket_id, file_id, _mimetype, is_head
        expected = (
            "2",
            "2023-04-19",
            "2023-04-19",
            "file.txt",
            "1",
            "3",
            "",
            "True",
        )
        content = file.read().rstrip()
        assert content == ",".join(expected)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Awards loader."""

from ...load.postgresql.bulk import PostgreSQLCopyLoad
from ...load.postgresql.bulk.generators import ExistingDataTableGenerator
from ..models.awards import Awards


class ExistingAwardsLoad(PostgreSQLCopyLoad):
    """Existing awards class for data loading."""

    def __init__(self, db_uri, data_dir, **kwargs):
        """Constructor."""
        super().__init__(
            db_uri=db_uri,
            table_generators=[ExistingDataTableGenerator(tables=[Awards], pks=[])],
            data_dir=data_dir,
            existing_data=True,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration OAuth2Server action stream."""

from .applications import (
    OAuthApplicationCreateAction,
    OAuthApplicationDeleteAction,
    OAuthApplicationUpdateAction,
)
from .linked_accounts import (
    OAuthGHDisconnectToken,
    OAuthLinkedAccountConnectAction,
    OAuthLinkedAccountDisconnectAction,
)
from .tokens import (
    OAuthServerTokenCreateAction,
    OAuthServerTokenDeleteAction,
    OAuthServerTokenUpdateAction,
)

__all__ = (
    "OAuthApplicationCreateAction",
    "OAuthApplicationDeleteAction",
    "OAuthApplicationUpdateAction",
    "OAuthGHDisconnectToken",
    "OAuthLinkedAccountConnectAction",
    "OAuthLinkedAccountDisconnectAction",
    "OAuthServerTokenCreateAction",
    "OAuthServerTokenDeleteAction",
    "OAuthServerTokenUpdateAction",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration OAuth stream."""

from .load import (
    OAuthClientCopyLoad,
    OAuthServerClientCopyLoad,
    OAuthServerTokenCopyLoad,
)
from .transform import OAuthServerClientTransform, OAuthServerTokenTransform

__all__ = (
    "OAuthClientCopyLoad",
    "OAuthServerClientCopyLoad",
    "OAuthServerTokenCopyLoad",
    "OAuthServerClientTransform",
    "OAuthServerTokenTransform",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration GitHub load module."""

from ...load.postgresql.bulk import PostgreSQLCopyLoad
from ...load.postgresql.bulk.generators import ExistingDataTableGenerator
from ..models.github import Repository, WebhookEvent
from .table_generator import ReleaseTableGenerator


class ExistingWebhookEventsCopyLoad(PostgreSQLCopyLoad):
    """Webhook events loading of existing data."""

    def __init__(self, db_uri, data_dir, **kwargs):
        """Constructor."""
        super().__init__(
            db_uri=db_uri,
            table_generators=[
                # it passes on prepare so entries wont have an effect on it
                ExistingDataTableGenerator(tables=[WebhookEvent], pks=[]),
            ],
            data_dir=data_dir,
            existing_data=True,
            # **kwargs
        )


class ExistingGitHubRepositoriesCopyLoad(PostgreSQLCopyLoad):
    """GitHub repository loading of existing data."""

    def __init__(self, db_uri, data_dir, **kwargs):
        """Constructor."""
        super().__init__(
            db_uri=db_uri,
            table_generators=[
                # it passes on prepare so entries wont have an effect on it
                ExistingDataTableGenerator(tables=[Repository], pks=[]),
            ],
            data_dir=data_dir,
            existing_data=True,
            # **kwargs
        )


class GitHubReleasesCopyLoad(PostgreSQLCopyLoad):
    """GitHub release loading of existing data."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(
            table_generators=[
                # this tg will treat the entries passing through the stream
                ReleaseTableGenerator(),
            ],
            **kwargs
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Requests transform tests."""

import dictdiffer

from invenio_rdm_migrator.streams.files import (
    FilesBucketEntry,
    FilesInstanceEntry,
    FilesObjectVersionEntry,
    FilesTransform,
)


class MockFilesTransform(FilesTransform):
    """Mock class for files data transformation."""

    def _bucket(self, entry):
        """Transform the bucket."""
        return {"id": "1"}

    def _object_version(self, entry):
        """Transform the object_version."""
        return {"id": "1"}

    def _file(self, entry):
        """Transform the file."""
        return {"id": "1"}


def test_files_transform():
    """Test the full files transformation."""
    result = MockFilesTransform()._transform({})
    expected = {
        "bucket": {"id": "1"},
        "object_version": {"id": "1"},
        "file": {"id": "1"},
    }
    assert not list(dictdiffer.diff(result, expected))


class MockFilesBucketEntry(FilesBucketEntry):
    """Transform a single mock files bucket entry."""

    def _id(self, entry):
        """Returns the file bucket ID."""
        return "1"

    def _created(self, entry):
        """Returns the creation date."""
        return "2023-04-19"

    def _updated(self, entry):
        """Returns the update date."""
        return "2023-04-19"

    def _default_location(self, entry):
        """Returns the file bucket default location."""
        return "1"

    def _default_storage_class(self, entry):
        """Returns the file bucket default storage class."""
        return "L"

    def _size(self, entry):
        """Returns the file bucket size."""
        return 1234

    def _quota_size(self, entry):
        """Returns the file bucket quota size."""
        return None

    def _max_file_size(self, entry):
        """Returns the file bucket max file size."""
        return None

    def _locked(self, entry):
        """Returns if the file bucket is locked."""
        return False

    def _deleted(self, entry):
        """Returns if the file bucket is deleted."""
        return False


def test_files_bucket_entry():
    """Test the files bucket entry transformation."""
    result = MockFilesBucketEntry().transform({})
    expected = {
        "id": "1",
        "created": "2023-04-19",
        "updated": "2023-04-19",
        "default_location": "1",
        "default_storage_class": "L",
        "size": 1234,
        "quota_size": None,
        "max_file_size": None,
        "locked": False,
        "deleted": False,
    }

    assert not list(dictdiffer.diff(result, expected))


class MockFilesObjectVersionEntry(FilesObjectVersionEntry):
    """Transform a single mock files bucket entry."""

    def _version_id(self, entry):
        """Returns the file object version ID."""
        return "2"

    def _created(self, entry):
        """Returns the creation date."""
        return "2023-04-19"

    def _updated(self, entry):
        """Returns the update date."""
        return "2023-04-19"

    def _key(self, entry):
        """Returns the file key name."""
        return "file.txt"

    def _bucket_id(self, entry):
        """Returns the file object version bucket ID i.e associated bucket."""
        return "1"

    def _file_id(self, entry):
        """Returns the file object version file ID i.e associated file."""
        return "3"

    def _mimetype(self, entry):
        """Returns the file object version mimetype."""
        return None

    def _is_head(self, entry):
        """Returns if the file object version is head i.e the latest."""
        return True


def test_files_object_version_entry():
    """Test the files object version entry transformation."""
    result = MockFilesObjectVersionEntry().transform({})
    expected = {
        "version_id": "2",
        "created": "2023-04-19",
        "updated": "2023-04-19",
        "key": "file.txt",
        "bucket_id": "1",
        "file_id": "3",
        "_mimetype": None,
        "is_head": True,
    }

    assert not list(dictdiffer.diff(result, expected))


class MockFilesInstanceEntry(FilesInstanceEntry):
    """Transform a single mock files bucket entry."""

    def _id(self, entry):
        """Returns the file instance ID."""
        return "3"

    def _created(self, entry):
        """Returns the creation date."""
        return "2023-04-19"

    def _updated(self, entry):
        """Returns the update date."""
        return "2023-04-19"

    def _uri(self, entry):
        """Returns the file uri."""
        return "path/to/file"

    def _storage_class(self, entry):
        """Returns the file instance storage class."""
        return "L"

    def _size(self, entry):
        """Returns the file instance size."""
        return 1234

    def _checksum(self, entry):
        """Returns the file instance checksum."""
        return "md5:abcd1234"

    def _readable(self, entry):
        """Returns if the file instance can be read."""
        return True

    def _writable(self, entry):
        """Returns if the file instance can be written."""
        return True

    def _last_check_at(self, entry):
        """Returns the last date the file was checked."""
        return None

    def _last_check(self, entry):
        """Returns if the file instance was last checked."""
        return False


def test_files_instance_entry():
    """Test the files instance entry transformation."""
    result = MockFilesInstanceEntry().transform({})
    expected = {
        "id": "3",
        "created": "2023-04-19",
        "updated": "2023-04-19",
        "uri": "path/to/file",
        "storage_class": "L",
        "size": 1234,
        "checksum": "md5:abcd1234",
        "readable": True,
        "writable": True,
        "last_check_at": None,
        "last_check": False,
    }

    assert not list(dictdiffer.diff(result, expected))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test utilities."""

import sqlalchemy as sa


def assert_model_count(session, model_cls, count):
    """Assert row count for a model."""
    assert session.scalar(sa.select(sa.func.count()).select_from(model_cls)) == count



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""OAuth2Server tokens actions module."""

from dataclasses import dataclass
from typing import Optional

from .....actions import LoadAction, LoadData
from .....load.postgresql.transactions.operations import Operation, OperationType
from ....models.oauth import ServerClient, ServerToken


@dataclass
class OAuthTokenData(LoadData):
    """OAuth server token action data."""

    client: Optional[dict] = None
    token: Optional[dict] = None


class OAuthServerTokenCreateAction(LoadAction):
    """Create a personal oauth token."""

    name = "oauth-server-token-create"
    data_cls = OAuthTokenData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new oauth token."""
        assert self.data.client and self.data.token

        yield Operation(OperationType.INSERT, ServerClient, self.data.client)
        yield Operation(OperationType.INSERT, ServerToken, self.data.token)


class OAuthServerTokenUpdateAction(LoadAction):
    """Create a personal oauth token."""

    name = "oauth-server-token-update"
    data_cls = OAuthTokenData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new oauth token."""
        if self.data.client:
            yield Operation(OperationType.UPDATE, ServerClient, self.data.client)
        if self.data.token:
            yield Operation(OperationType.UPDATE, ServerToken, self.data.token)


class OAuthServerTokenDeleteAction(LoadAction):
    """Create a personal oauth token."""

    name = "oauth-server-token-delete"
    data_cls = OAuthTokenData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new oauth token."""
        assert not self.data.client and self.data.token

        yield Operation(OperationType.DELETE, ServerToken, self.data.token)
